# -*- coding: utf-8 -*-
"""
Created on Wed Mar 23 16:25:29 2022

@author: Alexandre Coche Kenshilikov

Previously cwatres.py (changed on 20-07-2022)
Previously advancedresults.py (changed on 07-12-2022)
compare_maps() copied from comparison_maps.py
"""

#%% Imports :
# ~~~~~~~~~~~
import os
import pandas as pd
import geopandas as gpd
import numpy as np
import xarray as xr
xr.set_options(keep_attrs = True)
import rioxarray as rio #Not necessary, the rio module from xarray is enough
import rasterio
import rasterio.features
from affine import Affine
# from shapely.geometry import Point
import os
import sys
import re
# import matplotlib as mpl
import matplotlib.pyplot as plt
import readsettings as rs
from cwatm.management_modules.output import *

#%% WATERTABLE DEPTH
def watertable_depth(result_folder, timestep = 'daily', 
                     landcover = ['grassland', 'forest']):
    """
    DESCRIPTION :
    Calcule la profondeur de la nappe à partir de son altitude et de la
    topographie, en retranchant l'épaisseur du sol
    
    EXEMPLE :
    import visualization_advanced as va
    va.watertable_depth(r"D:\2- Postdoc\2- Travaux\3_CWatM_EBR\results\raw_results\005_calib_groundwater\08")
    
    # Obsolète
    va.watertable_depth(topo = r"D:\2- Postdoc\2- Travaux\3_CWatM_EBR\data\input_1km_LeMeu\ModFlow_inputs\75m\elevation_modflow.tif", 
                        watertable = r"D:\2- Postdoc\2- Travaux\3_CWatM_EBR\results\raw_results\005_calib_groundwater\01_base_Ronan\modflow_watertable_monthavg.nc", 
                        soil1 = r"D:\2- Postdoc\2- Travaux\3_CWatM_EBR\data\input_1km_LeMeu\soil\soildepth1_1km.nc", 
                        soil2 = r"D:\2- Postdoc\2- Travaux\3_CWatM_EBR\data\input_1km_LeMeu\soil\soildepth2_1km.nc", 
                        output = r"D:\2- Postdoc\2- Travaux\3_CWatM_EBR\results\raw_results\005_calib_groundwater\01_base_Ronan\modflow_watertable_depth_monthavg_QGIS.nc")
    
    To run for a whole screening
    ----------------------------
    for c in range(27, 31):
        for r in range(0, 616):
            print('\n-------')
            print(f'CAT = {c}    RUN = {r}')
            if not os.path.isfile(rf"D:\acoche\3_CWatM_EBR\results\raw_results\c{c:03n}\r{r:04n}\modflow_watertable_depth_daily.nc"):
                try:
                    va.watertable_depth(rf"D:\acoche\3_CWatM_EBR\results\raw_results\c{c:03n}\r{r:04n}", timestep = 'daily')
                except:
                    print('/!\ Failed')
            else:
                print('Already done')
    """
    
    
# =============================================================================
#     # TEMP INPUTS #
#     data_folder = r"D:\2- Postdoc\2- Travaux\3_CWatM_EBR\data\input_1km_LeMeu"
#     topo = os.path.join(data_folder, r"ModFlow_inputs\75m\elevation_modflow.tif")
#     result_folder = r"D:\2- Postdoc\2- Travaux\3_CWatM_EBR\results\raw_results\003_artif\2022-03-25_all_city"
#     watertable = os.path.join(result_folder, r"modflow_watertable_monthavg.nc")
#     soil1 = os.path.join(data_folder, r"soil\soildepth1_1km.nc")
#     soil2 = os.path.join(data_folder, r"soil\soildepth2_1km.nc")
#     output = os.path.join(result_folder, r"modflow_watertable_depth_monthavg_QGIS.nc")
#     mdfw_soil_thickness = os.path.join(result_folder, r"modflowtotalSoilThickness_totalend.nc")
# =============================================================================
    

    #%%% Chargement des données :
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~
    landcover_list = landcover
    if timestep == 'daily':
        decode_times_arg = True
    elif timestep == 'monthavg':
        decode_times_arg = False
    
    # Paramètres généraux
    # -------------------
    setting_file = [f for f in os.listdir(result_folder) if f.endswith('.ini')]
    if len(setting_file) == 1:
        setting_file = setting_file[0]
        (config, option, binding, outputDir) = rs.parse_configuration(
            os.path.join(result_folder, setting_file))
    else:
        print('\nNot able to find the setting file in the output_folder.')
        

    if result_folder[3:9] == 'acoche': # server Buffon
        cwatm_path = r"D:\acoche"
    elif result_folder[3:8] == 'Users': # server Risce
        # not implemented yet
        print('Server Risce: paths not standard')
        return
    else: # personal PC
        cwatm_path = r"D:\2- Postdoc\2- Travaux"
    main_folder_name = '3_CWatM_EBR'
    motif = re.compile(main_folder_name + '.*')
    print("\nRacine des fichiers d'input = {}".format(
        os.path.join(cwatm_path, main_folder_name)))
    
    # Création des chemins
    # --------------------
    mdfw_topo = os.path.join(cwatm_path, motif.search(binding['topo_modflow']).group())
    mdfw_river_percentage = os.path.join(cwatm_path, motif.search(binding['chanRatio']).group())
    mdfw_soil_thickness = os.path.join(result_folder, r"modflowtotalSoilThickness_totalend.nc")
    watertable = os.path.join(result_folder, f"modflow_watertable_{timestep}.nc")
    river_depth = os.path.join(cwatm_path, motif.search(binding['chanDepth']).group())
    output_file = os.path.join(result_folder, f"modflow_watertable_depth_{timestep}.nc")
    
    # For different landcovers
    # ------------------------
    path_rootDepth1 = dict()
    path_rootDepth2 = dict()
    path_rootDepth3 = dict()
    path_gravit_effsat1 = dict()
    path_gravit_effsat2 = dict()
    path_gravit_effsat3 = dict()
    path_fracVeg = dict()
    path_list = [path_rootDepth1, path_rootDepth2, path_rootDepth3,
                 path_gravit_effsat1, path_gravit_effsat2, path_gravit_effsat3, 
                 path_fracVeg,]
    for landcover in landcover_list:
        path_rootDepth1[landcover] = os.path.join(result_folder,
                                                  f"{landcover}_rootDepth1_totalend.nc")
        path_rootDepth2[landcover] = os.path.join(result_folder,
                                                  f"{landcover}_rootDepth2_totalend.nc")
        path_rootDepth3[landcover] = os.path.join(result_folder,
                                                  f"{landcover}_rootDepth3_totalend.nc")
        path_gravit_effsat1[landcover] = os.path.join(result_folder,
                                               f"{landcover}_gravit_effsat1_{timestep}.nc")
        path_gravit_effsat2[landcover] = os.path.join(result_folder,
                                               f"{landcover}_gravit_effsat2_{timestep}.nc")
        path_gravit_effsat3[landcover] = os.path.join(result_folder,
                                               f"{landcover}_gravit_effsat3_{timestep}.nc")
        path_fracVeg[landcover] = os.path.join(result_folder,
                                               f"frac{landcover.capitalize()}_totalend.nc")
    
    # Chargement MNT modflow
    # ----------------------
    with xr.open_dataset(mdfw_topo, 
                         decode_times = decode_times_arg) as topo_ds:
        topo_ds.load() # to unlock the resource
    topo_ds = topo_ds.where(topo_ds['band_data'] != -99999)  
        # car il y a visiblement dans le fichier initial de topo une redondance
        # des valeurs non affichées : nan et -99999, les dernières n'étant pas '_FillValue' ?
    topo_ds = topo_ds.sel(band = 1)
    topo_ds = topo_ds.drop_vars('band')
    
    # Chargement output watertable (réso modflow)
    # -------------------------------------------
    # NB : C'est une élévation absolue (= par rapprot niveau de la mer)
    with xr.open_dataset(watertable, 
                         decode_times = decode_times_arg, 
                         decode_coords = 'all') as watertable_ds:
        watertable_ds.load()
    
    main_var = list(watertable_ds.data_vars)[0] # par ex 'modflow_watertable_monthavg'
        
    # Alignment problem with Luca's modflow outputs
    if (watertable_ds['x'].values[0] != topo_ds.x.values[0]) | (watertable_ds['y'].values[0] != topo_ds.y.values[0]):
        print(f"\nL'emprise de l'output watertable est décalée de {(watertable_ds['x'].values[0]-topo_ds.x.values[0])/(watertable_ds['y'].values[1]-watertable_ds['y'].values[0])}x et {(watertable_ds['y'].values[0]-topo_ds.y.values[0])/(watertable_ds['y'].values[1]-watertable_ds['y'].values[0])}y")
        print("Réalignement effectué")
        watertable_ds['x'] = topo_ds.x
        watertable_ds['y'] = topo_ds.y
    

    # Chargement données sol (réso modflow)
    # -------------------------------------
    # with xr.open_dataset(soil1) as soil1_ds:
    #     soil1_ds.load()
    # with xr.open_dataset(soil2) as soil2_ds:
    #     soil2_ds.load()
    with xr.open_dataset(mdfw_soil_thickness,
                         decode_coords = 'all',
                         decode_times = decode_times_arg) as mdfw_soil_thickness_ds:
        mdfw_soil_thickness_ds.load()

        if (mdfw_soil_thickness_ds['x'].values[0] != topo_ds.x.values[0]) | (mdfw_soil_thickness_ds['y'].values[0] != topo_ds.y.values[0]):
            print(f"\nL'emprise de l'output mdfw_soil_thickness est décalée de {(mdfw_soil_thickness_ds['x'].values[0]-topo_ds.x.values[0])/(mdfw_soil_thickness_ds['y'].values[1]-mdfw_soil_thickness_ds['y'].values[0])}x et {(mdfw_soil_thickness_ds['y'].values[0]-topo_ds.y.values[0])/(mdfw_soil_thickness_ds['y'].values[1]-mdfw_soil_thickness_ds['y'].values[0])}y")
            print("Réalignement effectué")
            mdfw_soil_thickness_ds['x'] = topo_ds.x
            mdfw_soil_thickness_ds['y'] = topo_ds.y
    
    # Chargement couches de sol (réso modflow), pour différents landcovers
    # --------------------------------------------------
    rootDepth1_ds = xr.Dataset()
    rootDepth2_ds = xr.Dataset()
    rootDepth3_ds = xr.Dataset()
    gravit_effsat1_ds = xr.Dataset()
    gravit_effsat2_ds = xr.Dataset()
    gravit_effsat3_ds = xr.Dataset()
    fracVeg_ds = xr.Dataset()
    ds_list = [rootDepth1_ds, rootDepth2_ds, rootDepth3_ds,
               gravit_effsat1_ds, gravit_effsat2_ds, gravit_effsat3_ds, 
               fracVeg_ds,]
    for i in range(0, 7):
        for landcover in landcover_list:
            with xr.open_dataset(path_list[i][landcover],
                                 decode_coords = 'all',
                                 decode_times = decode_times_arg) as temp:
                ds_list[i][landcover] = temp[list(temp.data_vars)[0]]
        # Correction of attributes
        ds_list[i].attrs['standard_name'] = '_'.join(os.path.split(path_list[i]['grassland'])[-1].split('_')[-2:])[:-3]
        ds_list[i].attrs['long_name'] = ''

    # Chargement pourcentages de rivières (réso modflow)
    # --------------------------------------------------
    # Il est nécessaire d'appliquer une correction pour les rivières :
    with xr.open_dataset(mdfw_river_percentage,
                         decode_times = decode_times_arg) as mdfw_river_percentage_ds:
        mdfw_river_percentage_ds.load()
    mdfw_river_percentage_ds = mdfw_river_percentage_ds.drop('band')
    mdfw_river_percentage_ds = mdfw_river_percentage_ds.squeeze('band')
    
    
    # Récupération SCR cwatm
    # ----------------------
    maskmap = os.path.join(cwatm_path, motif.search(binding['MaskMap']).group())
    if os.path.splitext(maskmap)[-1] == ".nc":
        decode_coords_arg = 'all'
    else:
        decode_coords_arg = None

    with xr.open_dataset(maskmap, decode_coords = decode_coords_arg) as mask_ds:
        mask_ds.load()
    if 'spatial_ref' in list(mask_ds.coords) + list(mask_ds.data_vars):
        cwatm_epsg = mask_ds.rio.crs.to_epsg()
        print("\nCRS of CWatM is epsg:{}".format(cwatm_epsg))
    else:
        cwatm_epsg = 3035
        print("\nCRS of CWatM inputs is assumed to be epsg:{}".format(cwatm_epsg))
    
    mdfw_maskmap = os.path.join(cwatm_path, motif.search(binding['modflow_basin']).group())
    if os.path.splitext(mdfw_maskmap)[-1] == ".nc":
        decode_coords_arg = 'all'
    else:
        decode_coords_arg = None
        
    with xr.open_dataset(mdfw_maskmap, decode_coords = decode_coords_arg) as mdfw_mask_ds:
        mdfw_mask_ds.load()
    if 'spatial_ref' in list(mdfw_mask_ds.coords) + list(mdfw_mask_ds.data_vars):
        mdfw_epsg = mdfw_mask_ds.rio.crs.to_epsg()
        print("CRS of Modflow is epsg:{}".format(mdfw_epsg))
    else:
        mdfw_epsg = 2154
        print("CRS of Modflow inputs is assumed to be epsg:{}".format(mdfw_epsg))
    
    
# =============================================================================
#     # Chargement de la profondeur des rivières (réso CWatM)
#     # -----------------------------------------------------
#     """
#     La profondeur des rivières n'est finalement pas prise en compte par Modflow
#     """
#     with xr.open_dataset(river_depth) as river_depth_ds:
#         river_depth_ds.load()
#         river_depth_ds.rio.write_crs("epsg:{}".format(cwatm_epsg), inplace = True)
#     
#     # Reprojection :
#     river_depth_reprj = river_depth_ds.rio.reproject(
#         'epsg:{}'.format(mdfw_epsg), 
#         # resolution = (1000, 1000),
#         shape = (500, 650),
#         transform = Affine(75, 0.0, mdfw_river_percentage_ds.x.min().values-(75*10.5),
#                            0.0, -75, mdfw_river_percentage_ds.y.max().values+(75*10.5)), 
#         resampling = rasterio.enums.Resampling(5),
#         nodata = np.nan)
#     
#     
#     # Seules les cellules considérées comme rivières par modflow sont conservées :
#     mdfw_river_depth_ds = mdfw_river_percentage_ds.copy()
#     mdfw_river_depth_ds['band_data'] = mdfw_river_depth_ds.band_data * river_depth_reprj.chanbnkf
# =============================================================================
    
    
    #%%% Calcul basique de la profondeur de nappe
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# =============================================================================
# # Correction du sol
# Dans les settings, dans le § [GROUNDWATER_MODFLOW], on trouve :
# # Defining the upper limit of the groundwater layer
# use_soildepth_as_GWtop = False
# correct_soildepth_underlakes = False
# depth_underlakes = 1
# 
# Je pense que c'est de cette correction que Luca parle
# 
# On peut visualiser cette correction dans l'output modflowtotalSoilThickness_totalend.nc
# =============================================================================

    
    watertable_depth_ds = watertable_ds.copy()
    use_soildepth_as_GWtop = binding['use_soildepth_as_GWtop']
    
    #% 1| Profondeur de la nappe relativement à la topo :
    # --------------------------------------------------
    if use_soildepth_as_GWtop:
        """
        Cette formule donne la profondeur de la nappe par rapport à la topo,
        en considérant que la nappe ne peut pas empiéter sur le sol (la prof.
        de la nappe est au minimum égale à l'épaisseur du sol).
        La profondeur est donnée en valeurs positives.
        """
        watertable_depth_ds[main_var] = topo_ds['band_data'] \
            - watertable_ds[main_var]
        
    else:
        watertable_depth_ds[main_var] = topo_ds['band_data'] \
            - watertable_ds[main_var] \
                - mdfw_soil_thickness_ds['modflowtotalSoilThickness_totalend']
        
# =============================================================================
#     #% 2| Profondeur de la nappe relativement à la base du sol :
#     # ----------------------------------------------------------
#     """
#     Ou dit autrement : prof. de la nappe + du sol par rapport à la topo
#     """
#     if use_soildepth_as_GWtop:
#         watertable_depth_ds[main_var] = watertable_ds[
#             main_var] + mdfw_soil_thickness_ds[
#                 'modflowtotalSoilThickness_totalend'] - topo_ds['band_data']
#                     
#     else:
#         watertable_depth_ds[main_var] = watertable_ds[
#             main_var] - topo_ds['band_data']
# =============================================================================
        
# =============================================================================
#     # Il est possible de prendre en compte la profondeur des rivières
#     # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#     # Mais la prof. des rivières n'est pas prise en compte par Modflow, 
#     # seulement par CWatM pour le ROUTING
#     
#     #% Premier type de correction des rivières :
#     # ------------------------------------------
#     watertable_depth_ds['modflow_watertable_monthavg'] = watertable_depth_ds[
#         'modflow_watertable_monthavg'] + mdfw_river_depth_ds[
#             'band_data'] - topo_ds['band_data']    
#       
#     #% Deuxième type de correction des rivières :
#     # -------------------------------------------
#     watertable_depth_ds['modflow_watertable_monthavg'] = xr.where(
#         mdfw_river_percentage_ds.band_data == 1,
#         0,
#         watertable_depth_ds['modflow_watertable_monthavg'],
#         keep_attrs = False)
#     # Marche pas, je ne sais pas pourquoi...
# =============================================================================
    
# =============================================================================
#     #% Pour appliquer une profondeur nulle aux pixels avec des rivières
#     watertable_depth_ds[
#         main_var] = watertable_depth_ds.modflow_watertable_monthavg.where(
#         mdfw_river_percentage_ds.band_data == 0,
#         0)
# =============================================================================
    
# =============================================================================
#     # Enlever les pixels initiaux NaN ?
#     watertable_depth_ds[
#         main_var] = watertable_depth_ds.modflow_watertable_monthavg.where(
#         ~np.isnan(watertable_ds.modflow_watertable_monthavg.isel(time = 0)),
#         np.nan)
# =============================================================================
    
    #%%% Addition of water in the bottom layer of soil (subsoil)
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    watersheet3_ds = xr.Dataset()    # Dataset containing the equivalent water level [m] in 3rd soil layer (subsoil) to add to watertable
    watersheet3_reprj = xr.Dataset() # Previous dataset, reprojected
    fracVeg_reprj = xr.Dataset()     # Dataset containing landcover fractions, reprojected
    
    # Convert the effective saturation into equivalent watersheet
    for landcover in landcover_list:
        watersheet3_ds[landcover] = rootDepth3_ds[landcover] * gravit_effsat3_ds[landcover]
    
    # Reproject:
    reso_x = float(watertable_depth_ds.x[1] - watertable_depth_ds.x[0])
    reso_y = float(watertable_depth_ds.y[1] - watertable_depth_ds.y[0])
    transform_ = Affine(reso_x, 
                        0.0, 
                        float(watertable_depth_ds.x.min() - reso_x/2),  
                        0.0, 
                        reso_y, # reso_y is negative here
                        float(watertable_depth_ds.y.max() - reso_y/2))
    
    height_ = int(watertable_depth_ds.y.shape[0])
    width_ = int(watertable_depth_ds.x.shape[0])
    
    watersheet3_ds.rio.write_crs(cwatm_epsg, inplace = True)
    watersheet3_ds = watersheet3_ds.transpose('time', 'y', 'x')
    
    watersheet3_reprj = watersheet3_ds.rio.reproject(dst_crs = mdfw_epsg,
                                                     transform = transform_,
                                                     resampling = rasterio.enums.Resampling(5),
                                                     shape = (height_, width_),
                                                     # nodata = np.nan
                                                     )
    
    # Keep values only where watertable reaches the soil bottom (replace values with 0 everywhere else)
    for landcover in landcover_list:
        watersheet3_reprj[landcover] = xr.where(
            watertable_depth_ds[main_var] >= -mdfw_soil_thickness_ds['modflowtotalSoilThickness_totalend'],
            watersheet3_reprj[landcover],
            0)

        
    # Reproject landcover fractions
    fracVeg_ds.rio.write_crs(cwatm_epsg, inplace = True)
    fracVeg_reprj = fracVeg_ds.rio.reproject(dst_crs = mdfw_epsg,
                                             transform = transform_,
                                             resampling = rasterio.enums.Resampling(5),
                                             shape = (height_, width_),
                                             # nodata = np.nan
                                             )
        
    # Weighted average according to landcovers   
     # Initialization
    watersheet3_reprj['weighted'] = 0
    fracVeg_reprj['sum'] = 0
     # Loop
    for landcover in landcover_list:
        watersheet3_reprj['weighted'] = watersheet3_reprj['weighted'] + watersheet3_reprj[landcover] * fracVeg_reprj[landcover]
        fracVeg_reprj['sum'] = fracVeg_reprj['sum'] + fracVeg_reprj[landcover]
    watersheet3_reprj['normed_average'] = watersheet3_reprj['weighted'] / fracVeg_reprj['sum']
    
    # Add to watertable_depth_ds
    watertable_depth_ds[main_var] = watertable_depth_ds[main_var] + watersheet3_reprj['normed_average']


    #%%% Addition of water in the middle layer of soil (topsoil)
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    watersheet2_ds = xr.Dataset()    # Dataset containing the equivalent water level [m] in 2nd soil layer (bulk soil) to add to watertable
    watersheet2_reprj = xr.Dataset() # Previous dataset, reprojected
    # fracVeg_reprj_dict[ already created
    
    # Convert the effective saturation into equivalent watersheet
    for landcover in landcover_list:
        watersheet2_ds[landcover] = rootDepth2_ds[landcover] * gravit_effsat2_ds[landcover]
    
    # Reproject:   
    watersheet2_ds.rio.write_crs(cwatm_epsg, inplace = True)
    watersheet2_ds = watersheet2_ds.transpose('time', 'y', 'x')
    
    watersheet2_reprj = watersheet2_ds.rio.reproject(dst_crs = mdfw_epsg,
                                                     transform = transform_,
                                                     resampling = rasterio.enums.Resampling(5),
                                                     shape = (height_, width_),
                                                     # nodata = np.nan
                                                     )
    
    

# =============================================================================
#         middle_layer_limit = -0.3
# =============================================================================
    rootDepth12_ds = rootDepth1_ds + rootDepth2_ds
    # Reproject
    rootDepth12_ds.rio.write_crs(cwatm_epsg, inplace = True)
    rootDepth12_ds = rootDepth12_ds.transpose('y', 'x')
    rootDepth12_ds = rootDepth12_ds.rio.reproject(dst_crs = mdfw_epsg,
                                                  transform = transform_,
                                                  resampling = rasterio.enums.Resampling(5),
                                                  shape = (height_, width_),
                                                  # nodata = np.nan
                                                  )
    
    # Keep values only where below soil layer is saturated (replace values with 0 everywhere else)
    for landcover in landcover_list:
        watersheet2_reprj[landcover] = xr.where(
            watertable_depth_ds[main_var] >= -rootDepth12_ds[landcover],
            watersheet2_reprj[landcover],
            0,
            keep_attrs = False)

# =============================================================================
#     for landcover in landcover_list:
#         watersheet2_reprj[landcover] = xr.where(
#             watertable_depth_ds[main_var] >= -rootDepth12_ds[landcover],
#             watersheet2_reprj[landcover],
#             0,
#             keep_attrs = False)
# =============================================================================
    

    # Weighted average according to landcovers   
     # Initialization
    watersheet2_reprj['weighted'] = 0
    fracVeg_reprj['sum'] = 0
     # Loop
    for landcover in landcover_list:
        watersheet2_reprj['weighted'] = watersheet2_reprj['weighted'] + watersheet2_reprj[landcover] * fracVeg_reprj[landcover]
        fracVeg_reprj['sum'] = fracVeg_reprj['sum'] + fracVeg_reprj[landcover]
    watersheet2_reprj['normed_average'] = watersheet2_reprj['weighted'] / fracVeg_reprj['sum']
    
    # Add to watertable_depth_ds
    watertable_depth_ds[main_var] = watertable_depth_ds[main_var] + watersheet2_reprj['normed_average']

        
        # NB: modflow total soil thickness is 0.75% thicker than input
                                
    # watertable_depth_ds.where(
    #     watertable_depth_ds[
    #         main_var] >= -mdfw_soil_thickness_ds[
    #             'modflowtotalSoilThickness_totalend'],
    #     watertable_depth_ds[
    #         main_var] - np.mean(watersheet3_dict['grassland'].loc[dict(x = watertable_depth_ds.x, y = watertable_depth_ds.y)], 
    #                             # watersheet3_dict['forest'].loc[dict(x = watertable_depth_ds.x, y = watertable_depth_ds.y)]
    #                             ),
    #     drop = False
    #     )                                
            

    #%%% Formatage et export
    # ~~~~~~~~~~~~~~~~~~~~~~
    if timestep == 'daily':
        timestep_suffix = ''
    else:
        timestep_suffix = timestep
    wt_depth_var = 'modflow_watertable_depth' + timestep_suffix
    watertable_depth_ds = watertable_depth_ds.rename({main_var: wt_depth_var})
    watertable_depth_ds.x.attrs = {'standard_name': 'projection_x_coordinate',
                                   'long_name': 'x coordinate of projection',
                                   'units': 'Meter'}
    watertable_depth_ds.y.attrs = {'standard_name': 'projection_y_coordinate',
                                   'long_name': 'y coordinate of projection',
                                   'units': 'Meter'}
    watertable_depth_ds[wt_depth_var].attrs = {'long_name': 'watertable depth',
                                               'units': 'Meter'}

    # Transfert des encoding (pour conserver la taille sur disque)
    watertable_depth_ds[wt_depth_var].encoding = watertable_ds[main_var].encoding

    watertable_depth_ds.rio.write_crs(mdfw_epsg, inplace = True)    
    
    watertable_depth_ds.to_netcdf(output_file)
    
    
    #%%% Wetlands
    # ~~~~~~~~~~~
    
    # wetland_ds = watertable_depth_ds[main_var].where(
    #     watertable_depth_ds[main_var] > -0.3,
    #     drop = False)*0+1
    
    wetland_limit = -0.30
    print(f"\nWetlands are defined by lands saturated at {-wetland_limit*100} cm below surface")
    
    wetland_da = xr.where(
        watertable_depth_ds[wt_depth_var] > wetland_limit,
        watertable_depth_ds[wt_depth_var]*0+1,
        watertable_depth_ds[wt_depth_var]*0
        )
    
    wetland_ds = wetland_da.to_dataset()
    wetland_ds = wetland_ds.rename({wt_depth_var: 'wetland' + timestep_suffix})
    
    output_file_wl = os.path.join(result_folder, f"wetland_{timestep}.nc") 
    wetland_ds.x.attrs = {'standard_name': 'projection_x_coordinate',
                        'long_name': 'x coordinate of projection',
                        'units': 'Meter'}
    wetland_ds.y.attrs = {'standard_name': 'projection_y_coordinate',
                        'long_name': 'y coordinate of projection',
                        'units': 'Meter'}
    wetland_ds['wetland' + timestep_suffix].encoding = watertable_ds[main_var].encoding
    wetland_ds.rio.write_crs(mdfw_epsg, inplace = True)
    wetland_ds.to_netcdf(output_file_wl)
    
    # Average occurence
    mean_wetland_ds = wetland_ds.mean(dim = 'time')
    mean_wetland_ds = mean_wetland_ds.rename({'wetland' + timestep_suffix: 'wetland_totalavg'})
    
    output_file_mwl = os.path.join(result_folder, "wetland_totalavg.nc")
    mean_wetland_ds.x.attrs = {'standard_name': 'projection_x_coordinate',
                               'long_name': 'x coordinate of projection',
                               'units': 'Meter'}
    mean_wetland_ds.y.attrs = {'standard_name': 'projection_y_coordinate',
                               'long_name': 'y coordinate of projection',
                               'units': 'Meter'}
    mean_wetland_ds['wetland_totalavg'].attrs = {'long_name': 'proportion of wetlands',
                                                 }
    mean_wetland_ds['wetland_totalavg'].encoding = watertable_ds[main_var].encoding
    mean_wetland_ds.rio.write_crs(mdfw_epsg, inplace = True) 
    mean_wetland_ds.to_netcdf(output_file_mwl)
    
    
#%% WILTING POINT AND WATER STRESS
def dry(result_folder, landcover = 'grassland'):
    #%%% Chargement des données :
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Paramètres généraux
    # -------------------
    setting_file = [f for f in os.listdir(result_folder) if f.endswith('.ini')]
    if len(setting_file) == 1:
        setting_file = setting_file[0]
        (config, option, binding, outputDir) = rs.parse_configuration(
            os.path.join(result_folder, setting_file))
    else:
        print('\nNot able to find the setting file in the output_folder.')
        
    cwatm_path = r"D:\2- Postdoc\2- Travaux"
    main_folder_name = '3_CWatM_EBR'
    motif = re.compile(main_folder_name + '.*')
    print("\nRacine des fichiers d'input = {}".format(
        os.path.join(cwatm_path, main_folder_name)))
    
    # Création des chemins
    # --------------------
    grassland_w2 = os.path.join(result_folder, f"{landcover}_w2_daily.nc")
    grassland_wwp2 = os.path.join(result_folder, f"{landcover}_wwp2_totalend.nc")
    
# =============================================================================
#     grassland_rws = os.path.join(result_folder, r"grassland_rws_daily.nc")
# =============================================================================

    
    # Chargements
    # -----------
    with xr.open_dataset(grassland_w2, 
                         decode_times = True, 
                         decode_coords = 'all') as grassland_w2_ds:
        grassland_w2_ds.load()
    with xr.open_dataset(grassland_wwp2, 
                         # decode_times = True, 
                         decode_coords = 'all') as grassland_wwp2_ds:
        grassland_wwp2_ds.load()
        
# =============================================================================
#     with xr.open_dataset(grassland_rws, 
#                          decode_times = True, 
#                          decode_coords = 'all') as grassland_rws_ds:
#         grassland_rws_ds.load()
# =============================================================================
    
    
    #%%% Calculs
    # ~~~~~~~~~~
    wilted_ds = xr.where(
        grassland_w2_ds[f'{landcover}_w2'] < grassland_wwp2_ds[f'{landcover}_wwp2_totalend'],
        grassland_w2_ds[f'{landcover}_w2']*0+1,
        grassland_w2_ds[f'{landcover}_w2']*0
        )
    
    output_file_wilted = os.path.join(result_folder, "wilted_daily.nc")
    wilted_ds.rio.write_crs(3035, inplace = True)
    wilted_ds.to_netcdf(output_file_wilted)
    
    # Average occurence
    mean_wilted_ds = wilted_ds.mean(dim = 'time')
    
    output_file_mwilted = os.path.join(result_folder, "wilted_avg.nc")
    mean_wilted_ds.rio.write_crs(3035, inplace = True) 
    mean_wilted_ds.to_netcdf(output_file_mwilted)
    
# =============================================================================
#     # Water stress reduction factor
#     mean_grassland_rws_ds = grassland_rws_ds.mean(dim = 'time')
#     
#     output_file_mrws = os.path.join(result_folder, "grassland_rws_avg.nc")
#     mean_grassland_rws_ds.rio.write_crs(3035, inplace = True) 
#     mean_grassland_rws_ds.to_netcdf(output_file_mrws)
# =============================================================================
    
        
    
    #%%% Problèmes et notes
    # ~~~~~~~~~~~~~~~~~~~~~
    """
    • il faut corriger la profondeur des rivières [...]
    
    • la nappe se retrouve souvent au-dessus de 0 m (2-20 cm)
        → ça peut être une charge hydraulique 
    
    • parfois la nappe au-dessus du sol de 15 m ...
        → c'est dans les carrières (c'est négligeable)
                
    • L'activation de use_soildepth_as_GWtop ne change rien aux débits, recharges...
    """
    

#%% CREATE GRAVIT_EFFSAT_timestep.nc files (gravitationnal effective saturation)
def gravit_effsat(folder_path):
    """
    

    Parameters
    ----------
    folder_path : str
        Path to the folder containing outputs (or root path when
                                               recursive exploration will be
                                               implemented)

    Returns
    -------
    grassland_gravit_effsat_daily.nc etc. files are created in the same directory.

    """
    

    file_list = os.listdir(folder_path)
    
    #%%% Loading mandatory files
    # field capacity
    wfc1_ds = xr.Dataset()
    wfc2_ds = xr.Dataset()
    wfc3_ds = xr.Dataset()
    wfc_list = [wfc1_ds, wfc2_ds, wfc3_ds]
    # saturation water content
    ws1_ds = xr.Dataset()
    ws2_ds = xr.Dataset()
    ws3_ds = xr.Dataset()
    ws_list = [ws1_ds, ws2_ds, ws3_ds]
    for layer in range(0, 3):
        for landcover in ['forest', 'grassland']:
            # field capacity
            name = f'{landcover}_wfc{layer+1}_totalend.nc'
            if name in file_list:
                with xr.open_dataset(os.path.join(folder_path, name),
                                     decode_coords = 'all',
                                     decode_times = False) as temp:
                    wfc_list[layer][landcover] = temp[list(temp.data_vars)[0]]
            # saturation water content
            name = f'{landcover}_ws{layer+1}_totalend.nc'
            if name in file_list:
                with xr.open_dataset(os.path.join(folder_path, name),
                                     decode_coords = 'all',
                                     decode_times = False) as temp:
                    ws_list[layer][landcover] = temp[list(temp.data_vars)[0]]

    
    #%%% Exploring files
    for file_ in file_list:
        if os.path.splitext(file_)[-1] == '.nc': # if the element is a NetCDF file
            name = os.path.splitext(file_)[0]
            pattern_gravit = re.compile('.*gravit.*')
            pattern_sum = re.compile('.*sum.*')
            res_gravit = pattern_gravit.findall(name)
            res_sum = pattern_sum.findall(name)
            if (len(res_gravit) == 0) & (len(res_sum) == 0): # if filename contains neither 'sum' either 'gravit'
                pattern_effsat = re.compile('effsat\d*')
                int_el = pattern_effsat.findall(name)
                if len(int_el) > 0:
                    layer = int(int_el[0][-1]) - 1
                    ext_el = pattern_effsat.split(name)
                    new_name = 'gravit_' + int_el[0]
                    new_name = new_name.join(ext_el)
                    landcover = name.split('_')[0]
                    print(f"- File '{name}' as basis for '{new_name}'")
                    
                    #%%% Computation
                    soilwater_name = f'w{layer+1}'.join(ext_el)
                    with xr.open_dataset(os.path.join(folder_path, soilwater_name + '.nc'),
                                         decode_coords = 'all',
                                         decode_times = False) as data:
                        data.load() # to unlock the ressource
                    data = data.rename({list(data.data_vars)[0]: landcover}) # to have a consistent name for operations
                    data[landcover] = (data[landcover] - wfc_list[layer][landcover]) / (ws_list[layer][landcover] - wfc_list[layer][landcover])
                    # data = data.where(data[landcover] > 0, 0)
                    data = xr.where(data[landcover] < 0, 0, data)
                    
                    # Formatting
                    data = data.rename({landcover: new_name})
                    with xr.open_dataset(os.path.join(folder_path, name + '.nc'),
                                         decode_coords = 'all',
                                         decode_times = False) as effsat:
                        data[new_name].attrs = effsat[list(effsat.data_vars)[0]].attrs
                        data[new_name].encoding = effsat[list(effsat.data_vars)[0]].attrs
                    data.rio.write_crs(3035, inplace = True)
                    
                    # Exporting
                    data.to_netcdf(os.path.join(folder_path, new_name + '.nc'))
        
        
        elif os.path.isdir(os.path.join(folder_path, file_)):
            print('') # saut de ligne
            print(rf'Exploring {file_}\...') 
            gravit_effsat(os.path.join(folder_path, file_))
                
    

#%% ANIM QGIS
def date_counter(*, output_folder, start_date, fqce = 'M', **kwargs):
    """
    % DESCRIPTION :
    Exporte une série d'images *.png servant de compteur de dates pour
    l'animation qgis.
    
    % EXEMPLES :
    > import advancedresults as ar
    > ar.date_counter(output_folder = r"D:\2- Postdoc\2- Travaux\3_CWatM_EBR\results\processed_results\003\animation", 
                      start_date = '2000-01-01', fqce = 'M', period = 180)
    
    % ARGUMENTS :
    > start_date = date de départ
    > fqce = fréquence des pas de temps : 'D' | 'M' (défaut) | 'MS' (début du mois)
    > kwargs:
        > end_date = date de fin
            Si rien n'est spécifié, il faut qu'il y ait une 'period' de 
            spécifiée.
        > period = nombre de pas de temps.
            Si rien n'est spécifié, il faut qu'il y ait une 'end_date'
    """
    
    
    #% Get fields:
    # ------------
    if 'end_date' in kwargs:
        end_date = kwargs['end_date']
    else:
        end_date = None
    
    if 'period' in kwargs:
        period = kwargs['period']
    else:
        period = None
    
    
    #% Préparation de l'index :
    # -------------------------
    date_index = pd.date_range(start = start_date, 
                               end = end_date,  
                               periods = period, 
                               freq = fqce).strftime("%Y-%m-%d")
    
    
    #% Création et export des images :
    # --------------------------------
    for i, _date in enumerate(date_index):
        fig1, ax1 = plt.subplots(1, figsize = (3.5, 0.625)) 
        # inches à 96dpi = 336x60
        fig1.suptitle(_date, fontsize = 40)
        ax1.axis('off')
        filename = 'date_' + str(f'{i:04d}')
        fig1.savefig(os.path.join(output_folder, filename + '.png'), transparent = True)
        plt.close(fig1)
        
    return date_index    
    

#%% COMPARE NETCDF MAPS
def compare_maps(folder_ref, folder_comp, filename, mode):
    """
    Example
    -------
    va.compare_maps(folder_ref = r"D:\2- Postdoc\2- Travaux\3_CWatM_EBR\results\raw_results\017_new_data_landcover_meteo\01",
                      folder_comp = r"D:\2- Postdoc\2- Travaux\3_CWatM_EBR\results\raw_results\017_new_data_landcover_meteo\07",
                      filename = "sum_gwRecharge_monthavg.nc",
                      mode = "ratio")

    Parameters
    ----------
    folder_ref : str
        Name of the folder first to compare (reference).
    folder_comp : str
        Name of the folder to compare to the first folder.
    filename : str
        Name of the file to process.
    mode : str
        ratio | diff | relative difference.

    Returns
    -------
    Create a *.nc file in the folder "D:\2- Postdoc\2- Travaux\3_CWatM_EBR\results\processed_results"

    """
    
    #%% Load data
    # -----------
    print('\nLoading data...')
    # If the extension has not been included, include it:
    if os.path.splitext(filename)[-1] == '':
        print(' NB: Extension has not been indicated. Netcdf files will be considered.')
        filename = filename + '.nc'
        
    # Check if files exist. Otherwise show the closest name
    if not os.path.isfile(os.path.join(folder_ref, filename)) or not os.path.isfile(os.path.join(folder_comp, filename)):
        print(' WARNING: Issues with filename...')
        space_ref = [f for f in os.listdir(folder_ref) 
                     if os.path.isfile(os.path.join(folder_ref, f)) 
                     and os.path.splitext(f)[-1] == '.nc']
        space_comp = [f for f in os.listdir(folder_comp) 
                     if os.path.isfile(os.path.join(folder_comp, f)) 
                     and os.path.splitext(f)[-1] == '.nc']
        
        if not os.path.isfile(os.path.join(folder_ref, filename)):
            print('    File not found in first folder')
            closest_match = difflib.get_close_matches(filename, space_ref)
            if not closest_match: 
                common_match = ["- no match -"]
            else:  
                i = 0
                f = closest_match[i]
                while f not in space_comp and i < len(closest_match)-1:
                    i+=1
                    f = closest_match[i]
                common_match = f
            
        if not os.path.isfile(os.path.join(folder_comp, filename)):
            print('    File not found in second folder')
            closest_match = difflib.get_close_matches(filename, space_comp)
            if not closest_match: 
                common_match = ["- no match -"]
            else:  
                i = 0
                f = closest_match[i]
                while f not in space_ref and i < len(closest_match)-1:
                    i+=1
                    f = closest_match[i]
                common_match = f
            
        print('    Closest file to this name is "{}"'.format(common_match))
        filename = common_match
        print('    IMPORTANT: This file will be processed here')
    
    
    # Load datasets
    with xr.open_dataset(os.path.join(folder_ref, filename), 
                         decode_coords = 'all', 
                         decode_times = False) as data_ref:
        data_ref.load() # to unlock the resource
    
    with xr.open_dataset(os.path.join(folder_comp, filename), 
                         decode_coords = 'all', 
                         decode_times = False) as data_comp:
        data_comp.load() # to unlock the resource
    
    
    
    
    #%% Compute
    # ---------
    print('\nComputing...')
    if mode.casefold() == "ratio":
        mode_prefix = 'RATIO'
        res = data_comp/data_ref
        
    elif mode.casefold() in ['difference', 'diff']:
        mode_prefix = 'DIFF'
        res = data_comp-data_ref
        
    elif mode.casefold() in ['relative difference', 'rel diff', 'reldiff', 'diffrel']:
        mode_prefix = 'RELDIFF'
        res = (data_comp-data_ref)/data_ref
        
    
    #%% Export
    # --------
    print('\nExporting...')
    motif = re.compile('\d{3,3}.*\d{2,2}')
    extr_str_ref = motif.search(folder_ref).group()
    id_ref = extr_str_ref[0:3] + 'r' + os.path.split(folder_ref)[-1]
    extr_str_comp = motif.search(folder_comp).group()
    id_comp = extr_str_comp[0:3] + 'r' + os.path.split(folder_comp)[-1]
    output_name = os.path.join(r"D:\2- Postdoc\2- Travaux\3_CWatM_EBR\results\processed_results",
                               '_'.join([os.path.splitext(filename)[0], mode_prefix, id_comp, 'vs', id_ref]) + '.nc'
                               )
    
    res.to_netcdf(output_name)
    
    print(' Result file successfully created: \n {}'.format(output_name))
    
    return res
    

#%% GET FOLDERS : Internal function to get the current folder
def get_folders():
    script_folder = os.path.dirname(os.path.realpath(__file__))
    
    # Defining the root, result and settings folders (should be user-adjusted)
    root_folder = os.path.dirname(os.path.dirname(script_folder))
    result_folder = os.path.join(root_folder, 'results', 'raw_results')
    settings_design_folder = os.path.join(root_folder, 
                                          'data',
                                          'input_1km_LeMeu',
                                          'settings_and_experiment_designs',
                                          )
    
    return [script_folder, root_folder, result_folder, settings_design_folder]

    


# -*- coding: utf-8 -*-
"""
Created on Mon Jul 11 18:10:35 2022

Based on CWatM functions
"""

import configparser
import re
# import xml.dom.minidom
# It is necessary to add (...)\CWatM-version1.06 or (...)\CWATM_MODFLOW to the PYTHONPATH

import os

import difflib  # to check the closest word in settingsfile, if an error occurs


class ExtParser(configparser.ConfigParser):
    """
    addition to the parser to replace placeholders

    Example:
        PathRoot = C:/work
        MaskMap = $(FILE_PATHS:PathRoot)/data/areamaps/area.tif

    """

    #implementing extended interpolation
    def __init__(self, *args, **kwargs):
        self.cur_depth = 0
        configparser.ConfigParser.__init__(self, *args, **kwargs)

    def get(self, section, option, raw=False, vars=None, **kwargs):
        """
        def get(self, section, option, raw=False, vars=None
        placeholder replacement

        :param section: section part of the settings file
        :param option: option part of the settings file
        :param raw:
        :param vars:
        :return:
        """

        #h1 = sys.tracebacklimit
        #sys.tracebacklimit = 0  # no long error message
        try:
           r_opt = configparser.ConfigParser.get(self, section, option, raw=True, vars=vars)
        except:
             print(section, option)
             closest = difflib.get_close_matches(option, list(binding.keys()))
             if not closest: closest = ["- no match -"]
             #msg = "No key with the name: \"" + option +" in " + section + "\" in the settings file: \"" + settingsfile[0] + "\"\n"
             msg = "Error 116: Closest key to the required one is: \"" + closest[0] + "\""
             raise Exception(msg)


        #sys.tracebacklimit = h1   # set error message back to default
        if raw:
            return r_opt

        ret = r_opt
        re_newintp1 = r'\$\((\w*):(\w*)\)'  # other section
        re_newintp2 = r'\$\((\w*)\)'  # same section

        re_old1 = re.findall('\$\(\w*:\w*\)', r_opt)
        re_old2 = re.findall('\$\(\w*\)', r_opt)

        m_new1 = re.findall(re_newintp1, r_opt)
        m_new2 = re.findall(re_newintp2, r_opt)

        if m_new1:
             i = 0
             for f_section, f_option in m_new1:
                 self.cur_depth = self.cur_depth + 1
                 if self.cur_depth < configparser.MAX_INTERPOLATION_DEPTH:
                     sub = self.get(f_section,f_option, vars=vars)
                     ret = ret.replace(re_old1[i], sub)
                     i += 1
                 else:
                     raise configparser.InterpolationDepthError(option, section, r_opt)

        if m_new2:
             i = 0
             for l_option in m_new2:
                 self.cur_depth = self.cur_depth + 1
                 if self.cur_depth < configparser.MAX_INTERPOLATION_DEPTH:
                     sub = self.get(section, l_option, vars=vars)
                     ret = ret.replace(re_old2[i], sub)
                     i =+ 1
                 else:
                     raise configparser.InterpolationDepthError(option, section, r_opt)

        self.cur_depth = self.cur_depth - 1
        return ret



def parse_configuration(settingsFileName):
    """
    Parse settings file

    :param settingsFileName: name of the settings file
    :return: parameters in list: binding, options in list: option
    """

    # Initializations:
    # --------------
    option = {}
    binding = {}
    outputDir = []
    outDir = {}
    outMap = {}
    outTss = {}
    outsection = []
        
    def splitout(varin, check):
        """
        split variable in several one, seperator = ,

        :param varin:
        :param check:
        :return: list with sveral variables
        """

        out = list(map(str.strip, varin.split(',')))
        if out[0] == "": out[0]="None"
        if out[0] != "None": check = True
        return out, check

    if not(os.path.isfile(settingsFileName)):
        msg = "Error 302: Settingsfile not found!\n"
        raise Exception(settingsFileName,msg)
    config = ExtParser()
    config.optionxform = str
    config.sections()
    config.read(settingsFileName)
    for sec in config.sections():
        #print sec
        options = config.options(sec)
        check_section = False
        # option = dict.fromkeys(options) # Useless initialization of option
        for opt in options:
            if sec=="OPTIONS":
                try:
                    option[opt] = config.getboolean(sec, opt)
                except:
                    option[opt] = config.getint(sec, opt)
            else:
                # Check if config line = output line
                if opt.lower()[0:4] == "out_":
                    index = sec.lower()+"_"+opt.lower()

                    if opt.lower()[-4:] =="_dir":
                        outDir[sec] = config.get(sec, opt)
                    else:
                        # split into timeseries and maps
                        if opt.lower()[4:8] == "tss_":
                            outTss[index],check_section = splitout(config.get(sec, opt),check_section)
                        else:
                            outMap[index],check_section = splitout(config.get(sec, opt),check_section)

                else:
                    # binding: all the parameters which are not output or option are collected
                    binding[opt] = config.get(sec, opt)

        if check_section:
            outsection.append(sec)

    outputDir.append(binding["PathOut"])
     # Output directory is stored in a separat global array

    return(config, option, binding, outputDir)


# -*- coding: utf-8 -*-
"""
Created on Fri Dec 13 12:17:57 2024

@author: Don Kirkby
sources: https://stackoverflow.com/questions/552744/how-do-i-profile-memory-usage-in-python
and https://docs.python.org/3/library/tracemalloc.html
"""

#%% IMPORTS
import tracemalloc
from collections import Counter
import linecache
import os

#%% MEMORY MONITORING
def display_top(snapshot, key_type='lineno', limit=3):
    snapshot = snapshot.filter_traces((
        tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
        tracemalloc.Filter(False, "<unknown>"),
    ))
    top_stats = snapshot.statistics(key_type)
    
    print('\n----- MEMORY MONITORING -----')
    print("Top %s lines" % limit)
    for index, stat in enumerate(top_stats[:limit], 1):
        frame = stat.traceback[0]
        # replace "/path/to/module/file.py" with "module/file.py"
        # filename = os.sep.join(frame.filename.split(os.sep)[-2:])
        print("#%s: %s:%s: %.1f KiB"
              % (index, frame.filename, frame.lineno, stat.size / 1024))
        line = linecache.getline(frame.filename, frame.lineno).strip()
        if line:
            print('    %s' % line)

    other = top_stats[limit:]
    if other:
        size = sum(stat.size for stat in other)
        print("%s other: %.1f KiB" % (len(other), size / 1024))
    total = sum(stat.size for stat in top_stats)
    print("Total allocated size: %.1f KiB" % (total / 1024))
    print('-----------------------------\n')



#%% Imports
import geoconvert as gc
import os
import time

#%% Process ERA5-Land files
def process_era5(root_folder = r"D:\2- Postdoc\2- Travaux\1- Veille\4- Donnees\8- Meteo\ERA5\Brittany"):
    start_time = time.time()
    
    print("-----------------------------------")
    print("PREPROCESSING OF ERA5-LAND RAW DATA")
    print("-----------------------------------")
    #%% Initialization
    print("\n\n INITIALIZATION...")
    print(" ~~~~~~~~~~~~~~~~~")
    print(f'   Root folder: {root_folder}')
    
    date = ['1989-1999', '2000-2010', '2011-2021']
    print(f'   Dates: {date}')
    
    varname = ['Potential evaporation pos', # [m]/h
               'Relative humidity', # [%]
               'Surface pressure', # [Pa]
               'Surface solar radiation downwards W', # [W/m²]
               'Surface thermal radiation downwards W', # [W/m²]
               'Temperature', # [K]
               'Temperature',
               'Temperature',
               'Total precipitation', # [m]/h
               'Wind speed', # [m/s]
               ]
    
    mode = ['sum',
            'mean',
            'mean',
            'mean',
            'mean',
            'mean',
            'min',
            'max',
            'sum',
            'mean',
            ]


    #%% Update of variables
    print("\n\n UPDATING VARIABLES...")
    print(" ~~~~~~~~~~~~~~~~~~~~~")
    d=0
    for date_ in date:
        d+=1
    
    #-------------------------------------
    ### CREATE THE NECESSARY NEW VARIABLES
    #-------------------------------------
        #%%% Potential evaporations
        # gc.compute_Erefs_from_ERA5L(r"1989-1999 Potential evaporation.nc")
        print("\n    1- Switch potential evaporation from negative to positive")
        print("    __________________________________________________________")
        print(f"                                           date {d}/{len(date)}: {date_}")
        gc.compute_Erefs_from_ERA5L(
            os.path.join(root_folder, " ".join([date_, "Potential evaporation.nc"]))
            )
        #%%% Wind speed
        print("\n    2- Compute wind speed from Northing and Easting components")
        print("    ___________________________________________________________")
        print(f"                                            date {d}/{len(date)}: {date_}")
        gc.compute_wind_speed(
            os.path.join(root_folder, " ".join([date_, "U-component of wind.nc"]))
            )
    
        #%%% Relative humidity
        print("\n    3- Compute relative humidity from dewpoint, temperature and pressure")
        print("    _____________________________________________________________________")
        print(f"                                                      date {d}/{len(date)}: {date_}")
        gc.compute_relative_humidity(
            dewpoint_input_file = os.path.join(
                root_folder, " ".join([date_, "Dewpoint temperature.nc"])), 
            temperature_input_file = os.path.join(
                root_folder, " ".join([date_, "Temperature.nc"])),
            pressure_input_file = os.path.join(
                root_folder, " ".join([date_, "Surface pressure.nc"])),
            method = "Sonntag"
            )
    
    #----------------------------
    ### CONVERT UNITS WHEN NEEDED
    #----------------------------
        #%%% Radiations
        print("\n    4- Convert radiation units from J/m²/h to W/m²")
        print("    _______________________________________________")
        print(f"                              date {d}/{len(date)}: {date_}")
        gc.convert_downwards_radiation(os.path.join(
            root_folder, " ".join([date_, "Surface solar radiation downwards.nc"]))
            )
        gc.convert_downwards_radiation(os.path.join(
            root_folder, " ".join([date_, "Surface thermal radiation downwards.nc"]))
            )
    
    
    
        
    
    #%% Half-process checkup
    print("\n    > Half-process checkup <")
    print("    _________________________\n")
    print("\n".join(["\nAt this stage the folder should contain 8 (16) files per year:",
                     "    (Dewpoint temperature [K])",
                     "    (Potential evaporation crop)",
                     "  * Potential evaporation pos [m]",
                     "    (Potential evaporation water)",
                     "    (Potential evaporation)",
                     "  * Relative humidity [%]",
                     "  * Surface pressure [Pa]", 
                     "  * Surface solar radiation downwards W [W/m²]", 
                     "    (Surface solar radiation downwards [J/m²])",
                     "  * Surface thermal radiation downwards W [W/m²]",
                     "    (Surface thermal radiation downwards [J/m²])",
                     "  * Temperature [K]",
                     "  * Total precipitation [m]", 
                     "    (U-component of wind [m/s])",
                     "    (V-component of wind [m/s])",
                     "  * Wind speed [m/s]",
                     "",
                     "   (...) are original files not further needed"]))
        
    
    #%% Convert hourly into daily
    print("\n    5- Generate daily data from hourly data (in \data folder)")
    print("    __________________________________________________________")
    print("                                                     all dates\n")
    path_list = [os.path.join(
        root_folder, date_ + " " + varname[i] + ".nc") for date_ in date for i in range(0, len(varname))]
    
    mode_list = mode*len(date)
    
    # for date_ in date:
    #     for i in range(0, len(varname)):
    #         gc.hourly_to_daily(input_file = os.path.join(
    #             root_folder, date_ + " " + varname[i] + ".nc"), mode = mode[i])
    gc.hourly_to_daily(input_file = path_list, mode = mode_list)
    
    
    #%% Prepare final data files for CWatM
    print("\n    6- Generate ready-to-use data files for CWatM input (in \cwatm folder)")
    print("    _______________________________________________________________________")
    print("                                                                  all dates\n")
    for i in range(0, len(varname)):
        gc.prepare_CWatM_input(input_file = os.path.join(
            root_folder, "daily_temp", " ".join([date[0], varname[i], f"daily_{mode[i]}.nc"])),
            file_type = 'ERA5', reso_m = 6000, EPSG_out = 3035)
# =============================================================================
#     gc.prepare_CWatM_input(input_file = os.path.join(
#         root_folder, " ".join([date[0], "Potential evaporation water daily_mean.nc.nc"])),
#         file_type = 'ERA5', reso_m = 6000, CRS = 3035)
#     gc.prepare_CWatM_input(input_file = r"D:\2- Postdoc\2- Travaux\1- Veille\4- Donnees\8- Meteo\ERA5\Brittany\daily\1989-1999 Potential evaporation water daily_mean.nc", file_type = 'ERA5', reso_m = 6000, CRS = 3035)
#     gc.prepare_CWatM_input(input_file = r"D:\2- Postdoc\2- Travaux\1- Veille\4- Donnees\8- Meteo\ERA5\Brittany\daily\1989-1999 Relative humidity daily_mean.nc", file_type = 'ERA5', reso_m = 6000, CRS = 3035)
#     gc.prepare_CWatM_input(input_file = r"D:\2- Postdoc\2- Travaux\1- Veille\4- Donnees\8- Meteo\ERA5\Brittany\daily\1989-1999 Surface pressure daily_mean.nc", file_type = 'ERA5', reso_m = 6000, CRS = 3035)
#     gc.prepare_CWatM_input(input_file = r"D:\2- Postdoc\2- Travaux\1- Veille\4- Donnees\8- Meteo\ERA5\Brittany\daily\1989-1999 Surface solar radiation downwards W daily_mean.nc", file_type = 'ERA5', reso_m = 6000, CRS = 3035)
#     gc.prepare_CWatM_input(input_file = r"D:\2- Postdoc\2- Travaux\1- Veille\4- Donnees\8- Meteo\ERA5\Brittany\daily\1989-1999 Surface thermal radiation downwards W daily_mean.nc", file_type = 'ERA5', reso_m = 6000, CRS = 3035)
#     gc.prepare_CWatM_input(input_file = r"D:\2- Postdoc\2- Travaux\1- Veille\4- Donnees\8- Meteo\ERA5\Brittany\daily\1989-1999 Temperature daily_max.nc", file_type = 'ERA5', reso_m = 6000, CRS = 3035)
#     gc.prepare_CWatM_input(input_file = r"D:\2- Postdoc\2- Travaux\1- Veille\4- Donnees\8- Meteo\ERA5\Brittany\daily\1989-1999 Temperature daily_min.nc", file_type = 'ERA5', reso_m = 6000, CRS = 3035)
#     gc.prepare_CWatM_input(input_file = r"D:\2- Postdoc\2- Travaux\1- Veille\4- Donnees\8- Meteo\ERA5\Brittany\daily\1989-1999 Temperature daily_mean.nc", file_type = 'ERA5', reso_m = 6000, CRS = 3035)
#     gc.prepare_CWatM_input(input_file = r"D:\2- Postdoc\2- Travaux\1- Veille\4- Donnees\8- Meteo\ERA5\Brittany\daily\1989-1999 Total precipitation daily_mean.nc", file_type = 'ERA5', reso_m = 6000, CRS = 3035)
#     gc.prepare_CWatM_input(input_file = r"D:\2- Postdoc\2- Travaux\1- Veille\4- Donnees\8- Meteo\ERA5\Brittany\daily\1989-1999 Wind speed daily_mean.nc", file_type = 'ERA5', reso_m = 6000, CRS = 3035)
#     
# =============================================================================
    
    print("\n\n ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
    print(f" Elapsed time: {time.time() - start_time} s")    


#%% Correct ERA5-Land data
def correct_era5(input_folder = r"D:\2- Postdoc\2- Travaux\3_CWatM_EBR\data\input_1km_LeMeu\meteo\ERA5Land"):
    start_time = time.time()
    
    print("--------------------------------")
    print("CORRECTION OF ERA5-LAND RAW DATA")
    print("--------------------------------")
    #%% Initialization
    print("\n\n INITIALIZATION...")
    print(" ~~~~~~~~~~~~~~~~~")
    print(f'   Input folder: {input_folder}')
    
    date = ['1989-1999', '2000-2010', '2011-2021']
    print(f'   Dates: {date}')
    
    #%% Correct value
    print("\n\n CORRECTING VALUES...")
    print(" ~~~~~~~~~~~~~~~~~~~~")
    #%%% Precipitations
    print("\n    1- Multiplying precipitations by 0.087")
    print("    _______________________________________\n")
    gc.correct_era5(input_file = os.path.join(
        input_folder, r"1989-2021 Total precipitation res=6000m epsg3035.nc"),
        correct_factor = 0.087,
        to_dailysum = False)
    
    #%%% Downwards radiations
    print("\n    2- Multiplying solar radiations (W) by 0.0715")
    print("    ______________________________________________\n")
    gc.correct_era5(input_file = os.path.join(
        input_folder, r"1989-2021 SSRD Watt res=6000m epsg3035.nc"),
        correct_factor = 0.0715,
        to_dailysum = False,
        progressive = True)
    
    print("\n    3- Multiplying thermal radiations (W) by 0.0715")
    print("    ________________________________________________\n")
    gc.correct_era5(input_file = os.path.join(
        input_folder, r"1989-2021 STRD Watt res=6000m epsg3035.nc"),
        correct_factor = 0.0715,
        to_dailysum = False,
        progressive = True)
    
    #%%% Potential evapotranspiration
    print("\n    4- Multiplying potential evaporation by 0.04")
    print("    _____________________________________________\n")
    gc.correct_era5(input_file = os.path.join(
        input_folder, r"1989-2021 Potential evaporation pos res=6000m epsg3035.nc"),
        correct_factor = 0.04,
        to_dailysum = False)
    
    print("\n\n ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
    print(f" Elapsed time: {time.time() - start_time} s")   
    


# -*- coding: utf-8 -*-
"""
Created on Tue Sep 17 16:58:10 2024

@author: Alexandre Kenshilik Coche
@contact: alexandre.co@hotmail.fr

download.py est un ensemble de fonctions permettant le téléchargement
automatique de données françaises nécessaires à la modélisation 
hydrologique des socio-écosystèmes.
Cet outil a notamment vocation à regrouper les données classiques nécessaires
au modèle CWatM utilisé dans le cadre de la 
méthodologie Eau et Territoire (https://eau-et-territoire.org ).
"""

#%% IMPORTATIONS
import os
import re
import requests
import datetime
from io import (BytesIO, StringIO)
import gzip
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point
import xarray as xr
xr.set_options(keep_attrs = True)
from geop4th import geobricks as geo
from geop4th.workflows.standardize import standardize_fr as stzfr


#%% Utilitaries to check if a request is valid
def valid_request(response, 
                  file_format = 'geojson', 
                  prev_frame = None):
    """
    Utilitary to check if a request is valid.

    Parameters
    ----------
    response : request.response
        `request` object that contains a server's response to an HTTP request.
        This ``response`` variable can be obtained with ``requests.get(<url>)``
    file_format : {'csv', 'json', 'geojson'}, default 'geojson'
        File format in which the data will be retrieved.
    prev_frame : pandas.DataFrame or geopandas.GeoDataFrame, optional
        Frame used for recurrence run (when request results are on several pages)

    Returns
    -------
    isvalid : bool
        Flag indicating if the request has succeeded.
    frame : pandas.DataFrame of geopandas.GeoDataFrame
        Frame contaning the downloaded values.
        If ``file_format`` is 'geojson', returns a geopandas.GeoDataFrame, otherwise
        return a pandas.DataFrame (non-geospatialized data).

    """
    isvalid = False
    frame = prev_frame
    
    if (response.status_code == 200): # all correct, one single page
        _, frame = valid_single_request(response, file_format)
    
    elif (response.status_code == 206): # correct but results are on several pages
        isvalid, frame = valid_single_request(response, file_format)
        
        # Safeguard preventing technical limitations of BNPE API (max request size)
        # BNPE API can not handle more than 20000 results in one request
        if 'last' in response.links:
            page_pattern = re.compile("page=(\d*)")
            n_page = page_pattern.findall(response.links['last']['url'])
            if len(n_page) > 0:
                n_results = int(n_page[0]) * frame.shape[0]
                if n_results > 20000:
                    print(f"\nErr: Up to {n_results} results were asked but BNPE API cannot retrieve more than 20000 results in one request. If you passed a `masks` argument, please provide smaller masks.")
                    return False, None
        
        if prev_frame is not None:
            frame = pd.concat([prev_frame, frame], axis = 0, ignore_index = True).drop_duplicates()
        
        if 'next' in response.links:
            response = requests.get(response.links['next']['url'])
            _, frame = valid_request(response, file_format, 
                                           prev_frame = frame)
        
    elif (response.status_code == 400):
        print("Err: Incorrect request")
    elif (response.status_code == 401):
        print("Err: Unauthorized")
    elif (response.status_code == 403):
        print("Err: Forbidden")
    elif (response.status_code == 404):
        print("Err: Not Found")
    elif (response.status_code == 500):
        print("Err: Server internal error")
    
    # if frame is not None
    if isinstance(frame, pd.DataFrame):
        isvalid = True
        
    return isvalid, frame
        

def valid_single_request(response, 
                         file_format = 'geojson'):
    """
    Utilitary to check if a request is valid.

    Parameters
    ----------
    response : request.response
        `request` object that contains a server's response to an HTTP request.
        This ``response`` variable can be obtained with ``requests.get(<url>)``
    file_format : {'csv', 'json', 'geojson'}, default 'geojson'
        File format in which the data will be retrieved.

    Returns
    -------
    isvalid : bool
        Flag indicating if the request has succeeded.
    frame : pandas.DataFrame of geopandas.GeoDataFrame
        Frame contaning the downloaded values.
        If ``file_format`` is 'geojson', returns a geopandas.GeoDataFrame, otherwise
        return a pandas.DataFrame (non-geospatialized data).

    """
    
    # ---- Determine validity of downloaded data
    isvalid = False
    if file_format == 'csv':
        if response.text != '': # not empty
            isvalid = True
    elif file_format in ['json', 'geojson']:
        if response.json()['count'] != 0: # not empty
            isvalid = True
    
    if isvalid:  
        # ---- Retrieve results as a dataframe 
        # Save data content into a DataFrame (or a GeoDataFrame)
        if file_format == 'csv':
            frame = pd.read_csv(BytesIO(response.content), sep=";")
        elif file_format == 'geojson':
            # For some datasets, the 'GeoJSON' format option is not available,
            # only classic JSON is retrieved
            # In these cases, if the user requires a GeoJSON, the JSON dict
            # needs to be converted to GeoJSON before export :
                
            if 'type' in response.json():
                if response.json()['type'] == 'FeatureCollection':
                    # Response content is truly a GeoJSON
                    json_to_geojson = False
                else:
                    json_to_geojson = True
            
            else:
                json_to_geojson = True
            
            if json_to_geojson:
                # Additional step of georefencing are necessary    
                json_df = pd.DataFrame.from_dict(response.json()['data'])
                geometry = [Point(xy) for xy in zip(json_df.longitude, 
                                                    json_df.latitude)]
                frame = gpd.GeoDataFrame(
                    json_df,
                    crs = 4326,
                    geometry = geometry)
                
            else:
                frame = gpd.read_file(response.text, driver='GeoJSON')
            
        elif file_format == 'json':
# =============== useless =====================================================
#                     data_stations = pd.DataFrame.from_dict(
#                         {i: response.json()['data'][i] for i in range(len(response.json()['data']))},
#                         orient = 'index')
# =============================================================================
            frame = pd.DataFrame.from_dict(response.json()['data'])                
        
    else:
        frame = None
    
    return isvalid, frame


#%% BNPE (Données de prélèvements)
def bnpe(*, dst_folder, 
         start_year = None, 
         end_year = None, 
         masks = None, 
         departments = None, 
         communes = None, 
         file_formats = 'geojson'):
    """
    Function to facilitate the downloading of French water withdrawal data from BNPE API.

    Parameters
    ----------
    dst_folder : str or pathlib.Path
        Destination folder in which the downloaded data will be stored.
    start_year : int, optional
        Year from which the data will be retrieved.
    end_year : int, optional
        Year until which the data will be retrieved.
    masks : list of-, or single element among str, pathlib.Path, xarray.Dataset, xarray.DataArray or geopandas.GeoDataFrame, optional
        Mask on which the data will be retrieved.
        At least one parameter among ``masks``, ``departments`` or ``communes``
        should be passed.
    departments : int or list of int, optional
        INSEE code(s) of department(s) for which data will be retrieved.
        At least one parameter among ``masks``, ``departments`` or ``communes``
        should be passed.
    communes : int or list of int, optional
        INSEE code(s) of commune(s) for which data will be retrieved.
        At least one parameter among ``masks``, ``departments`` or ``communes``
        should be passed.
    file_formats : str or list of str, {'csv', 'json', 'geojson'}, default 'geojson'
        File format in which the data will be retrieved.

    Returns
    -------
    None. Downloaded data is stored in ``dst_folder``

    """
    
    # ---- Argument retrieving    
    # Safeguard
    if (departments is None) & (communes is None) & (masks is None):
        print("Err: It is required to specify at least one area with the arguments `departments` or `communes` or `masks`")
        return
    
    if start_year is None:
        start_year = 2008
        if end_year is None:
            end_year = datetime.datetime.today().year
    if isinstance(start_year, (str, float)):
        start_year = int(start_year)
    
    if end_year is None: # and start_year is not None, see previous case
        years = [start_year]
    else:
        if isinstance(end_year, (str, float)):
            end_year = int(end_year)
        years = list(range(start_year, end_year + 1))
    
    if masks is not None:
        if isinstance(masks, tuple):
            masks = list(masks)
        elif not isinstance(masks, list):
            masks = [masks]
    
    if departments is not None:
        if isinstance(departments, (str, float)):
            departments = [int(departments)]
        elif isinstance(departments, int):
            departments = [departments]
        elif isinstance(departments, tuple):
            departments = list(departments)
    
    if communes is not None:
        if isinstance(communes, (str, float)):
            communes = [int(communes)]
        elif isinstance(communes, int):
            communes = [communes]
        elif isinstance(communes, tuple):
            communes = list(communes)
    
    if isinstance(file_formats, str):
        file_formats = [file_formats]
    for i in range(0, len(file_formats)):
        # file_formats = file_formats.replace('.', '')
        if file_formats[i][0] == '.': file_formats[i] = file_formats[i][1:]
    
    outdir = os.path.join(dst_folder, "originaux")
    if not os.path.exists(outdir): os.makedirs(outdir)
    
    outdir_ouvrages = os.path.join(dst_folder, "originaux", "ouvrages")
    if not os.path.exists(outdir_ouvrages): os.makedirs(outdir_ouvrages)
    
    # ---- Requests
    print("\nDownloading...")
    for y in years:
        y = int(y)
        print(f"\n   . {y}")
        
        if departments is not None:
            for d in departments:
                d = f"{d:02.0f}" # format '00'
                
                for f in file_formats:
                    if f == 'csv':
                        url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/chroniques.csv?annee={y}&code_departement={d}&size=5000"
                        outpath = os.path.join(outdir, f"dpmt{d}_{y}.csv")
                    elif f == 'geojson':
                        url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/chroniques?annee={y}&code_departement={d}&format=geojson&size=5000"
                        outpath = os.path.join(outdir, f"dpmt{d}_{y}.json")
                    elif f == 'json':
                        url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/chroniques?annee={y}&code_departement={d}&size=5000"
                        outpath = os.path.join(outdir, f"dpmt{d}_{y}.json")
                    
                    response = requests.get(url)
                    
                    isvalid, frame = valid_request(response, f)
                    if isvalid:
                        # Export
                        if f == 'csv':
                            geo.export(frame, outpath, sep = ';', encoding = 'utf-8', index = False)
                        else:
                            geo.export(frame, outpath)
                                
        
        if communes is not None:
            for c in communes:
                c = f"{c:<05.0f}" # format '000000'
                
                for f in file_formats:
                    if f == 'csv':
                        url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/chroniques.csv?annee={y}&code_commune_insee={c}&size=5000"
                        outpath = os.path.join(outdir, f"cmne{c}_{y}.csv")
                    elif f == 'geojson':
                        url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/chroniques?annee={y}&code_commune_insee={c}&format=geojson&size=5000"
                        outpath = os.path.join(outdir, f"cmne{c}_{y}.json")
                    elif f == 'json':
                        url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/chroniques?annee={y}&code_commune_insee={c}&size=5000"
                        outpath = os.path.join(outdir, f"cmne{c}_{y}.json")
                    
                    response = requests.get(url)
            
                    isvalid, frame = valid_request(response, f)
                    if isvalid:
                        # Export
                        if f == 'csv':
                            geo.export(frame, outpath, sep = ';', encoding = 'utf-8', index = False)
                        else:
                            geo.export(frame, outpath)
                            
        if masks is not None:
            i_mask = 0
            for m in masks:
                i_mask += 1
                mask_ds = geo.load_any(m)
                mask_ds = geo.reproject(mask_ds, dst_crs = 4326) 
                
                for f in file_formats:
                    if f == 'csv':
                        url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/chroniques.csv?annee={y}&bbox={mask_ds.total_bounds[0]}&bbox={mask_ds.total_bounds[1]}&bbox={mask_ds.total_bounds[2]}&bbox={mask_ds.total_bounds[3]}&size=5000"
                        outpath = os.path.join(outdir, f"mask{i_mask}_{y}.csv")
                    elif f == 'geojson':
                        url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/chroniques?annee={y}&bbox={mask_ds.total_bounds[0]}&bbox={mask_ds.total_bounds[1]}&bbox={mask_ds.total_bounds[2]}&bbox={mask_ds.total_bounds[3]}&format=geojson&size=5000"
                        outpath = os.path.join(outdir, f"mask{i_mask}_{y}.json")
                    elif f == 'json':
                        url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/chroniques?annee={y}&bbox={mask_ds.total_bounds[0]}&bbox={mask_ds.total_bounds[1]}&bbox={mask_ds.total_bounds[2]}&bbox={mask_ds.total_bounds[3]}&size=5000"
                        outpath = os.path.join(outdir, f"mask{i_mask}_{y}.json")
                    
                    response = requests.get(url)
            
                    isvalid, frame = valid_request(response, f)
                    
# ======= {draft} to automatically handle too large requests ==================
#                     isvalid, frame, exceeds = valid_request(response, f)
#                     
#                     if exceeds:
#                         masks = [*masks[0:i_mask-1], split(masks[i_mask-1]), *masks[i_mask:]]
#                         bnpe(dst_folder = dst_folder, start_year = start_year, 
#                              end_year = end_year, masks = masks, 
#                                  departments = departments, communes = communes, 
#                                  file_formats = file_formats)
#                         return
# =============================================================================
                    
                    if isvalid:
                        # In the `mask` case, the data is clipped to the mask
                        if isinstance(frame, gpd.GeoDataFrame):
                            frame = geo.reproject(frame, mask = m)
                        else: # frame is only a pd.DataFrame
                            # first frame is converted to a GeoDataFrame
                            geometry = [Point(xy) for xy in zip(frame.longitude, 
                                                                frame.latitude)]
                            gdf = gpd.GeoDataFrame(
                                frame,
                                crs = 4326,
                                geometry = geometry)
                            # Then it is clipped
                            gdf = geo.reproject(gdf, mask = m)
                            # Finally it is converted back to a DataFrame
                            frame = pd.DataFrame(gdf.drop(columns = 'geometry'))

                        # Export
                        if f == 'csv':
                            geo.export(frame, outpath, sep = ';', encoding = 'utf-8', index = False)
                        else:
                            geo.export(frame, outpath)
                
    # Infos sur ouvrage
    if departments is not None:
        for d in departments:
            d = f"{d:02.0f}" # format '00'

            for f in file_formats:
                if f == 'csv':
                    url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/referentiel/ouvrages.csv?code_departement={d}&size=5000"
                    outpath = os.path.join(outdir_ouvrages, f"dpmt{d}_ouvrages.csv")
                elif f == 'geojson':
                    url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/referentiel/ouvrages?code_departement={d}&format=geojson&size=5000"
                    outpath = os.path.join(outdir_ouvrages, f"dpmt{d}_ouvrages.json")
                elif f == 'json':
                    url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/referentiel/ouvrages?code_departement={d}&size=5000"
                    outpath = os.path.join(outdir_ouvrages, f"dpmt{d}_ouvrages.json")
                
                response = requests.get(url)
                
                isvalid, frame = valid_request(response, f)
                if isvalid:
                    # Export
                    if f == 'csv':
                        geo.export(frame, outpath, sep = ';', encoding = 'utf-8', index = False)
                    else:
                        geo.export(frame, outpath)
        
    if communes is not None:
        for c in communes:
            c = f"{c:<05.0f}" # format '000000'
            
            for f in file_formats:
                if f == 'csv':
                    url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/referentiel/ouvrages.csv?code_commune_insee={c}&size=5000"
                    outpath = os.path.join(outdir_ouvrages, f"cmne{c}_ouvrages.csv")
                elif f == 'geojson':
                    url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/referentiel/ouvrages?code_commune_insee={c}&format=geojson&size=5000"
                    outpath = os.path.join(outdir_ouvrages, f"cmne{c}_ouvrages.json")
                elif f == 'json':
                    url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/referentiel/ouvrages?code_commune_insee={c}&size=5000"
                    outpath = os.path.join(outdir_ouvrages, f"cmne{c}_ouvrages.json")
                
                response = requests.get(url)
                
                isvalid, frame = valid_request(response, f)
                if isvalid:
                    # Export
                    if f == 'csv':
                        geo.export(frame, outpath, sep = ';', encoding = 'utf-8', index = False)
                    else:
                        geo.export(frame, outpath)
                        
    if masks is not None:
        i_mask = 0
        for m in masks:
            i_mask += 1
            mask_ds = geo.load_any(m)
            mask_ds = geo.reproject(mask_ds, dst_crs = 4326) 
            
            for f in file_formats:
                if f == 'csv':
                    url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/referentiel/ouvrages.csv?bbox={mask_ds.total_bounds[0]}&bbox={mask_ds.total_bounds[1]}&bbox={mask_ds.total_bounds[2]}&bbox={mask_ds.total_bounds[3]}&size=5000"
                    outpath = os.path.join(outdir_ouvrages, f"mask{i_mask}_ouvrages.csv")
                elif f == 'geojson':
                    url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/referentiel/ouvrages?bbox={mask_ds.total_bounds[0]}&bbox={mask_ds.total_bounds[1]}&bbox={mask_ds.total_bounds[2]}&bbox={mask_ds.total_bounds[3]}&format=geojson&size=5000"
                    outpath = os.path.join(outdir_ouvrages, f"mask{i_mask}_ouvrages.json")
                elif f == 'json':
                    url = rf"https://hubeau.eaufrance.fr/api/v1/prelevements/referentiel/ouvrages?bbox={mask_ds.total_bounds[0]}&bbox={mask_ds.total_bounds[1]}&bbox={mask_ds.total_bounds[2]}&bbox={mask_ds.total_bounds[3]}&size=5000"
                    outpath = os.path.join(outdir_ouvrages, f"mask{i_mask}_ouvrages.json")
                
                response = requests.get(url)
                
                isvalid, frame = valid_request(response, f)
                
                if isvalid:
                    # In the `mask` case, the data is clipped to the mask
                    if isinstance(frame, gpd.GeoDataFrame):
                        frame = geo.reproject(frame, mask = m)
                    else: # frame is only a pd.DataFrame
                        # first frame is converted to a GeoDataFrame
                        geometry = [Point(xy) for xy in zip(frame.longitude, 
                                                            frame.latitude)]
                        gdf = gpd.GeoDataFrame(
                            frame,
                            crs = 4326,
                            geometry = geometry)
                        # Then it is clipped
                        gdf = geo.reproject(gdf, mask = m)
                        # Finally it is converted back to a DataFrame
                        frame = pd.DataFrame(gdf.drop(columns = 'geometry'))
                
                    # Export
                    if f == 'csv':
                        geo.export(frame, outpath, sep = ';', encoding = 'utf-8', index = False)
                    else:
                        geo.export(frame, outpath)
                        

#%% Hydrometry (Mesures de débits)
def hydrometry(*, dst_folder, 
               masks, 
               start_year, 
               end_year = None, 
               quantities = 'QmnJ', 
               file_formats = 'geojson'):
    """
    Function to facilitate the downloading of French hydrometry data from EauFrance API.

    Parameters
    ----------
    dst_folder : str or pathlib.Path
        Destination folder in which the downloaded data will be stored.
    masks : list of-, or single element among str, pathlib.Path, xarray.Dataset, xarray.DataArray or geopandas.GeoDataFrame, optional
        Mask on which the data will be retrieved.
    start_year : int, optional
        Year from which the data will be retrieved.
    end_year : int, optional
        Year until which the data will be retrieved.
    quantities : str or list of str, default 'QmnJ'
    
        - 'QmnJ': average daily discharge
        - 'QmM': average monthly discharge
        - 'HIXM': maximum instant height per month
        - 'HIXnJ': maximum instant height per day
        - 'QINM': minimum instant discharge per month
        - 'QINnJ': minimum instant discharge per day
        - 'QixM': maximum instant discharge per month
        - 'QIXnJ': maximum instant discharge per day
    
    file_formats : str or list of str, {'csv', 'json', 'geojson'}, default 'geojson'
        File format in which the data will be retrieved.

    Returns
    -------
    None. Downloaded data is stored in ``dst_folder``
    """

    # ---- Argument retrieving
    if isinstance(start_year, (str, float)):
        start_year = int(start_year)
    if end_year is None:
        years = [start_year]
    else:
        if isinstance(end_year, (str, float)):
            end_year = int(end_year)
        years = list(range(start_year, end_year + 1))
    
    if masks is not None:
        if isinstance(masks, tuple):
            masks = list(masks)
        elif not isinstance(masks, list):
            masks = [masks]
    
    if isinstance(file_formats, str):
        file_formats = [file_formats]
    for i in range(0, len(file_formats)):
        # file_formats = file_formats.replace('.', '')
        if file_formats[i][0] == '.': file_formats[i] = file_formats[i][1:]
    
    if isinstance(quantities, str):
        quantities = [quantities]
    
    outdir = os.path.join(dst_folder, "originaux")
    if not os.path.exists(outdir): os.makedirs(outdir)
    
    outdir_stations = os.path.join(dst_folder, "originaux", "stations")
    if not os.path.exists(outdir_stations): os.makedirs(outdir_stations)

    # ---- Requests
    print("\nDownloading...")    
                           
    i_mask = 0
    for m in masks:
        i_mask += 1
        mask_ds = geo.load_any(m)
        mask_ds = geo.reproject(mask_ds, dst_crs = 4326) 
        
        # List of stations 
        for f in file_formats:
            if f == 'csv':
                url = rf"https://hubeau.eaufrance.fr/api/v2/hydrometrie/referentiel/stations.csv?bbox={mask_ds.total_bounds[0]}&bbox={mask_ds.total_bounds[1]}&bbox={mask_ds.total_bounds[2]}&bbox={mask_ds.total_bounds[3]}&size=10000"
                outpath = os.path.join(outdir_stations, f"mask{i_mask}_stations.csv")
            elif f == 'geojson':
                url = url = rf"https://hubeau.eaufrance.fr/api/v2/hydrometrie/referentiel/stations?bbox={mask_ds.total_bounds[0]}&bbox={mask_ds.total_bounds[1]}&bbox={mask_ds.total_bounds[2]}&bbox={mask_ds.total_bounds[3]}&format=geojson&size=10000"
                outpath = os.path.join(outdir_stations, f"mask{i_mask}_stations.json")
            elif f == 'json':
                url = url = rf"https://hubeau.eaufrance.fr/api/v2/hydrometrie/referentiel/stations?bbox={mask_ds.total_bounds[0]}&bbox={mask_ds.total_bounds[1]}&bbox={mask_ds.total_bounds[2]}&bbox={mask_ds.total_bounds[3]}&size=10000"
                outpath = os.path.join(outdir_stations, f"mask{i_mask}_stations.json")
            
            response = requests.get(url)
            
            isvalid, data_stations = valid_request(response, f)
            
            if isvalid:
                # Data is clipped to the mask
                if isinstance(data_stations, gpd.GeoDataFrame):
                    data_stations = geo.reproject(data_stations, mask = m)
                else: # frame is only a pd.DataFrame
                    # first frame is converted to a GeoDataFrame
                    data_stations.loc[:, ['longitude', 'latitude']] = data_stations['coordLatLon'].str.split(',', expand = True).rename(columns = {0: 'latitude', 1: 'longitude'}).astype(float)
                    geometry = [Point(xy) for xy in zip(data_stations.longitude, 
                                                        data_stations.latitude)]
                    gdf = gpd.GeoDataFrame(
                        data_stations,
                        crs = 4326,
                        geometry = geometry)
                    # Then it is clipped
                    gdf = geo.reproject(gdf, mask = m)
                    # Finally it is converted back to a DataFrame
                    data_stations = pd.DataFrame(gdf.drop(columns = ['geometry', 'latitude', 'longitude']))
                
                # Export
                if f == 'csv':
                    geo.export(data_stations, outpath, sep = ';', encoding = 'utf-8', index = False)
                else:
                    geo.export(data_stations, outpath)
            
        # Discharge data
        for y in years:
            y = int(y)
            print(f"   . {y}")
            
            for station_id in data_stations['code_station']:

                for f in file_formats:
                    for q in quantities:
                        if f == 'csv':                    
                            url = rf"https://hubeau.eaufrance.fr/api/v2/hydrometrie/obs_elab.csv?code_entite={station_id}&bbox={mask_ds.total_bounds[0]}&bbox={mask_ds.total_bounds[1]}&bbox={mask_ds.total_bounds[2]}&bbox={mask_ds.total_bounds[3]}&date_debut_obs_elab={y}-01-01&date_fin_obs_elab={y}-12-31&grandeur_hydro_elab={q}&size=20000"
                            outpath = os.path.join(outdir, f"{q}_{station_id}_mask{i_mask}_{y}.csv")
# ============ apparently format option is not available ======================
#                         elif f == 'geojson':
#                             url = rf"https://hubeau.eaufrance.fr/api/v2/hydrometrie/obs_elab?code_entite={station_id}&bbox={mask_ds.total_bounds[0]}&bbox={mask_ds.total_bounds[1]}&bbox={mask_ds.total_bounds[2]}&bbox={mask_ds.total_bounds[3]}&date_debut_obs_elab={y}-01-01&date_fin_obs_elab={y}-12-31&grandeur_hydro_elab={q}&format=geojson&size=20000"
#                             outpath = os.path.join(outdir, f"{q}_{station_id}_mask{i_mask}_{y}.json")
# =============================================================================
                        elif f in ['geojson', 'json']:
                            url = rf"https://hubeau.eaufrance.fr/api/v2/hydrometrie/obs_elab?code_entite={station_id}&bbox={mask_ds.total_bounds[0]}&bbox={mask_ds.total_bounds[1]}&bbox={mask_ds.total_bounds[2]}&bbox={mask_ds.total_bounds[3]}&date_debut_obs_elab={y}-01-01&date_fin_obs_elab={y}-12-31&grandeur_hydro_elab={q}&size=20000"
                            outpath = os.path.join(outdir, f"{q}_{station_id}_mask{i_mask}_{y}.json")
                        
                        response = requests.get(url)
                        
                        isvalid, frame = valid_request(response, f)
                        
                        if isvalid:
                            # Data is clipped to the mask
                            if isinstance(frame, gpd.GeoDataFrame):
                                frame = geo.reproject(frame, mask = m)
                            else: # frame is only a pd.DataFrame
                                # first frame is converted to a GeoDataFrame
                                geometry = [Point(xy) for xy in zip(frame.longitude, 
                                                                    frame.latitude)]
                                gdf = gpd.GeoDataFrame(
                                    frame,
                                    crs = 4326,
                                    geometry = geometry)
                                # Then it is clipped
                                gdf = geo.reproject(gdf, mask = m)
                                # Finally it is converted back to a DataFrame
                                frame = pd.DataFrame(gdf.drop(columns = 'geometry'))
                        
                            # Export
                            if f == 'csv':
                                geo.export(frame, outpath, sep = ';', encoding = 'utf-8', index = False)
                            else:
                                geo.export(frame, outpath)



#%% SIM2 (Données climatiques de réanalyse historique MétéoFrance)
def sim2(*, var_list, 
         dst_folder: str,
         first_year: int=None, 
         last_year: int=None,
         extension = '.nc', 
         mask: str=None):
    """
    Depuis le 13/12/2023 ces données sont en accès ouvert sur https://meteo.data.gouv.fr
    (onglet "Données climatologiques de référence pour le changement climatique",
    puis "Données changement climatique - SIM quotidienne" : 
    https://meteo.data.gouv.fr/datasets/6569b27598256cc583c917a7 )
    
    Parameters
    ----------
    var_list : iterable
        List of variable names.
        Works with SIM2 variable names: 'DRAINC_Q' | 'RUNC_Q' | 'EVAP_Q' | 'PRETOT_Q' | 'T_Q' ...
        The '_Q' is optional ('DRAINC', 'RUNC'... also work)
    dst_folder : str or pathlib.Path
        Path to the output folder containing the clipped SIM2 netCDF files.
        According to the files already present in this folder, only the necessary
        complementary files will be downloaded.
    first_year : int, optional
        Data will be extracted from 1st January of first_year to 31st December of last_year.
        If None, the earliest available date will be considered.
    last_year : int, optional
        End date of data to be extracted. 
        If None, the current date of the day will be used instead.
    extension : {'.nc', '.csv'}, default '.nc'
        Output file type.
    mask : str or pathlib.Path, optional
        Shapefile path to indicate how to clip
        the CSV or netCDF files that will be exported to the ``dst_folder`` folder.
        Note: The only purpose of using ``mask`` at this step is to save space on disk.

    Returns
    -------
    None. Create or update the netcdf files.
    """

    # ---- Initialization
    print("\nInitializing...")
    # Variables
# =============================================================================
#     if ('PRETOT' in var_list) | ('PRETOT_Q' in var_list):
#         var_list = var_list + ['PRENEI', 'PRELIQ']
# =============================================================================
    for i in range(0, len(var_list)):
        if var_list[i][-2:] != '_Q':
            var_list[i] = var_list[i] + '_Q'
    var_sublist = []
    
    # Folders
    if not os.path.isdir(dst_folder):
        os.makedirs(dst_folder)
    if extension[0] != '.': extension = '.' + extension
    
    # Dates
    if first_year is None:
        first_year = 1958
    first_date = pd.to_datetime(f"{first_year}-01-01", format = "%Y-%m-%d")
    
    if last_year is None:
        last_date = pd.to_datetime('today').normalize()
        last_year = last_date.year
    else:
        last_date = pd.to_datetime(f"{last_year}-12-31", format = "%Y-%m-%d")
    
# =============================================================================
#     # Mask
#     if mask is not None:
#         if not os.path.splitext(mask)[-1] == '.shp':
#             print("Err: The mask value should point to a .shp file. Otherwise, use mask=None to deactivate clipping")
#             return
# =============================================================================
    
    france_extent = (56000.0, 1613000.0, 1200000.0, 2685000.0) # whole France 
                                                               # limits for SIM2 
                                                               # data in epsg:2154 

    data_to_merge = []

    # ---- Determine which data needs to be downloaded
    # Data already available for each variable 

    local_data = pd.DataFrame(index = var_list, columns = ['file',
                                                           'start_date',
                                                           'end_date',
                                                           'extent'])
    local_data.extent = False

    sim_pattern = re.compile('(.*)_SIM2_')
    _, filelist = geo.get_filelist(dst_folder, extension = extension)
# =============================================================================
#     filelist = [os.path.join(dst_folder, f) \
#                 for f in os.listdir(dst_folder) \
#                     if os.path.isfile(os.path.join(dst_folder, f))]
# =============================================================================
    if len(filelist) > 0: # folder is not empty
        for file in filelist:
            filename = os.path.split(file)[-1]
            sim2_match = sim_pattern.findall(filename)
            var = sim2_match[0].replace("_QUOT", "")
            local_data.loc[var, 'file'] = file
            
            remove_file = False
            if extension == '.csv':
                df_temp = pd.read_csv(os.path.join(dst_folder, file), sep = ';', 
# =============================================================================
#                                       usecols = ['LAMBX', 'LAMBY', 'DATE'] + var_list,
# =============================================================================
                                      header = 0, decimal = '.',
                                      parse_dates = ['DATE'],
                                      # date_format='%Y%m%d', # Not available before pandas 2.0.0
                                      )
                
                # Variables
                if len(set(var_list).intersection(set(df_temp.columns))) < len(var_list):
                    print(f"   Local data {os.path.split(file)[-1]} does not include all desired variables")
                    remove_file = True # the file will be deleted (outside this section)
                
                # Dates
                if pd.date_range(start = df_temp.DATE.iloc[0], 
                                 end = df_temp.DATE.iloc[-1], 
                                 freq = 'D').size == df_temp.DATE.size: # all time values are contiguous
                    local_data.loc[var, 'start_date'] = pd.to_datetime(df_temp.DATE.iloc[0])
                    local_data.loc[var, 'end_date'] = pd.to_datetime(df_temp.DATE.iloc[-1])
            
            elif extension == '.nc':
                with geo.load_any(os.path.join(dst_folder, file), decode_coords = 'all', decode_times = True) as ds_temp:
                    # Dates
                    if pd.date_range(start = ds_temp.time[0].item(), 
                                     end = ds_temp.time[-1].item(), 
                                     freq = 'D').size == ds_temp.time.size: # all time values are contiguous
                        local_data.loc[var, 'start_date'] = pd.to_datetime(ds_temp.time[0].item())
                        local_data.loc[var, 'end_date'] = pd.to_datetime(ds_temp.time[-1].item())
                    # Spatial extent
                    resolution = abs(ds_temp.rio.resolution()[0])
                    ds_extent = np.array(ds_temp.rio.bounds())
                    if mask is not None:
                        mask_gdf = geo.load_any(mask)
                        mask_gdf = geo.reproject(mask_gdf, dst_crs = ds_temp.rio.crs)
                        # mask_extent = mask_gdf.buffer(resolution).total_bounds
                        if isinstance(mask_gdf, gpd.GeoDataFrame):
                            mask_extent = mask_gdf.total_bounds
                        elif isinstance(mask_gdf, xr.Dataset):
                            mask_extent = mask_gdf.rio.bounds()
                        mask_extent[0] = geo.nearest(
                            x = mask_extent[0], x0 = ds_extent[0], 
                            y0 = ds_extent[1], res = resolution)
                        if mask_extent[0] > mask_gdf.total_bounds[0]:
                            mask_extent[0] -= resolution
                        mask_extent[1] = geo.nearest(
                            y = mask_extent[1], x0 = ds_extent[0],
                            y0 = ds_extent[1], res = resolution)
                        if mask_extent[1] > mask_gdf.total_bounds[1]:
                            mask_extent[1] -= resolution
                        mask_extent[2] = geo.nearest(
                            x = mask_extent[2], x0 = ds_extent[0],
                            y0 = ds_extent[1], res = resolution)
                        if mask_extent[2] < mask_gdf.total_bounds[2]:
                            mask_extent[2] += resolution
                        mask_extent[3] = geo.nearest(
                            y = mask_extent[3], x0 = ds_extent[0],
                            y0 = ds_extent[1], res = resolution)
                        if mask_extent[3] < mask_gdf.total_bounds[3]:
                            mask_extent[3] += resolution
                            
                    else:
                        mask_extent = france_extent # whole France
                    
                    if (ds_extent[0:2] <= mask_extent[0:2]).any() | (ds_extent[2:4] >= mask_extent[2:4]).any() :
                        local_data.loc[var, 'extent'] = True
                    else:
                        print(f"   Local data {os.path.split(file)[-1]} does not cover desired spatial extent")
                        remove_file = True # the file will be deleted (outside this 'with' section)
            
            if remove_file:
                os.remove(os.path.join(dst_folder, file))
                   
        # local_data.iloc[:, 0:3][local_data.extent == True] = np.nan
        local_data.loc[local_data.index[local_data.extent == False], local_data.columns[0:3]] = np.nan
        
    
    # ---- Download
    print("\nDownloading...")
    # Download only the necessary data files from MeteoFrance API
    # Until the access to SIM2 data is implemented through the Météo-France's
    # API (https://portail-api.meteofrance.fr), the current stable urls
    # are used.
    
    stable_urls = {
        'QUOT_SIM2_1958-1959': ('https://www.data.gouv.fr/fr/datasets/r/5dfb33b3-fae5-4d0e-882d-7db74142bcae', 
                                0.16, pd.to_datetime('1958-08-01', format = "%Y-%m-%d"),
                                pd.to_datetime('1959-12-31', format = "%Y-%m-%d")),
        'QUOT_SIM2_1960-1969': ('https://www.data.gouv.fr/fr/datasets/r/eb0d6e42-cee6-4d7c-bc5b-646be4ced72e', 
                                1.1, pd.to_datetime('1960-01-01', format = "%Y-%m-%d"),
                                pd.to_datetime('1969-12-31', format = "%Y-%m-%d")),
        'QUOT_SIM2_1970-1979': ('https://www.data.gouv.fr/fr/datasets/r/33417617-c0dd-4513-804e-c3f563cb81b4', 
                                1.1, pd.to_datetime('1970-01-01', format = "%Y-%m-%d"),
                                pd.to_datetime('1979-12-31', format = "%Y-%m-%d")),
        'QUOT_SIM2_1980-1989': ('https://www.data.gouv.fr/fr/datasets/r/08ad5936-cb9e-4284-a6fc-36b29aca9607', 
                                1.1, pd.to_datetime('1980-01-01', format = "%Y-%m-%d"),
                                pd.to_datetime('1989-12-31', format = "%Y-%m-%d")),
        'QUOT_SIM2_1990-1999': ('https://www.data.gouv.fr/fr/datasets/r/ad584d65-7d2d-4ff1-bc63-4f93357ed196', 
                                1.1, pd.to_datetime('1990-01-01', format = "%Y-%m-%d"),
                                pd.to_datetime('1999-12-31', format = "%Y-%m-%d")),
        'QUOT_SIM2_2000-2009': ('https://www.data.gouv.fr/fr/datasets/r/10d2ce77-5c3b-44f8-bb46-4df27ed48595', 
                                1.1, pd.to_datetime('2000-01-01', format = "%Y-%m-%d"),
                                pd.to_datetime('2009-12-31', format = "%Y-%m-%d")),
        'QUOT_SIM2_2010-2019': ('https://www.data.gouv.fr/fr/datasets/r/da6cd598-498b-4e39-96ea-fae89a4a8a46', 
                                1.1, pd.to_datetime('2010-01-01', format = "%Y-%m-%d"),
                                pd.to_datetime('2019-12-31', format = "%Y-%m-%d")),
        'QUOT_SIM2_latest_period': ('https://www.data.gouv.fr/fr/datasets/r/92065ec0-ea6f-4f5e-8827-4344179c0a7f', 
                                    1.1, pd.to_datetime('2020-01-01', format = "%Y-%m-%d"),
                                    (pd.to_datetime('today').normalize() - pd.Timedelta(1, 'D')).replace(day = 1) - pd.Timedelta(1, 'D')),
        'QUOT_SIM2_latest_days': (#'https://www.data.gouv.fr/fr/datasets/r/ff8e9fc6-d269-45e8-a3c3-a738195ea92a',
                                  'https://www.data.gouv.fr/fr/datasets/r/adcca99a-6db0-495a-869f-40c888174a57',
                                   0.1, (pd.to_datetime('today').normalize() - pd.Timedelta(1, 'D')).replace(day = 1),
                                   pd.to_datetime('today').normalize() - pd.Timedelta(1, 'D')),
        }
    
    available_data = pd.DataFrame.from_dict(
        data = stable_urls, 
        orient = 'index', 
        columns = ['url', 'size_Go', 'start_date', 'end_date'])
    
    ### Identify which files will be needed
    # -------------------------------------
    # If there is no netcdf file already
    if local_data.file.isnull().values.any():
        # Then all the data files will be downloaded
        to_download = available_data.index[
            (available_data.end_date > first_date) \
                 & (available_data.start_date < last_date)]
    else:
        # The "core period" is the period that is covered by local data for 
        # specified variables
        min_core_date = local_data.start_date[var_list].max()
        max_core_date = local_data.end_date[var_list].min()
        
        if first_date < min_core_date:
            inf_idx = available_data.index[(available_data.start_date < min_core_date) & (available_data.end_date > first_date)]
        else:
            inf_idx = []
        
        if last_date > max_core_date:
            sup_idx = available_data.index[(available_data.end_date > max_core_date) & (available_data.start_date < last_date)]
        else:
            sup_idx = []
        
        to_download = list(set(inf_idx).union(set(sup_idx)))
    
    # Files to download to cover the times
    # (Note that if the specified spatial extent is not covered, the files have been previously deleted in __init__())
    if len(to_download) > 0:
        # print(f"The following .csv datasets will be downoladed: {', '.join([dataname + '(' + self.available_data.loc[dataname, 'size_Go'] + ')' for dataname in to_download])}")
        ram_space = available_data.loc[to_download, 'size_Go'].max()
        disk_space = available_data.loc[to_download, 'size_Go'].sum()/2.5*len(var_list) \
            - sum(os.path.getsize(os.path.join(dst_folder, f)) \
                  for f in os.listdir(dst_folder) \
                      if os.path.isfile(os.path.join(dst_folder, f)))/1073741824 
        disk_space = np.max([0, disk_space])
        print(f"The following CSV datasets will be downloaded to RAM and exported to netCDF files to cover the required period and area: {', '.join(to_download)}")
        print(f"(required RAM: < {ram_space} Go, required space: < {disk_space:.2f} Go)")
      
    ### Download the required files
    # -----------------------------
    if len(to_download) > 0:
        for dataname in to_download: 
            print(f"\nDownloading {dataname}...")
            print("   (can take several min, depending on internet speed)")
            response = requests.get(available_data.loc[dataname, 'url'])
    
            if (response.status_code == 200) & (os.path.splitext(response.url)[-1] != '.tmp'):
                # Decompress gzip content
                with gzip.open(BytesIO(response.content), 'rt') as f:
                    # Determine variables to extract in the current file
                    var_sublist = local_data.loc[var_list].index[
                        (local_data.start_date[var_list] > available_data.loc[dataname, 'start_date']) \
                            | (local_data.end_date[var_list] < available_data.loc[dataname, 'end_date'])]
                    var_sublist = var_sublist.to_list() + local_data[local_data.file.isnull()].index.to_list()  
                    # Replace 'PRETOT_Q' with 'PRELIQ_Q' and 'PRENEI_Q' if needed
                    var_list_csv = []
                    for v in var_list:
                        if v != 'PRETOT_Q':
                            var_list_csv += [v]
                        else:
                            var_list_csv += ['PRELIQ_Q', 'PRENEI_Q']
                    # Read .csv file and export desired variables
                    df = pd.read_csv(f, sep = ';', 
                                     usecols = ['LAMBX', 'LAMBY', 'DATE'] + var_list_csv,
                                     header = 0, decimal = '.',
                                     parse_dates = ['DATE'],
                                     # date_format='%Y%m%d', # Not available before pandas 2.0.0
                                     )
                    df = df[(df.DATE >= first_date) & (df.DATE <= last_date)]
                    
                    # Cumulated values (day 1) are summed from 06:00 UTC (day 1) to 06:00 UTC (day 2)
                    # Therefore, days correspond to Central Standard Time days.
# =============================================================================
#                     df['DATE'] = df['DATE'].tz_localize('Etc/GMT-6')
# =============================================================================

                    if 'PRETOT_Q' in var_list:
                        df['PRETOT_Q'] = df['PRELIQ_Q'] + df['PRENEI_Q']
                        if not 'PRELIQ_Q' in var_list:
                            df.drop(columns = 'PRELIQ_Q', inplace = True)
                        if not 'PRENEI_Q' in var_list:
                            df.drop(columns = 'PRENEI_Q', inplace = True)
                    
                    if extension == '.csv':
                        # Standardization step necessary for further clipping 
                        # (this step is realized in stzfr.sim2(df) when extension == '.nc' (see below))
                        df.rename(columns = {'LAMBX': 'x', 'LAMBY': 'y', 'DATE': 'time'}, inplace = True)
                        df[['x', 'y']] = df[['x', 'y']]*100 # convert hm to m
                        data_to_merge.append(df)
                    
                    elif extension == '.nc':
                        ds = stzfr.sim2(df)
                        data_to_merge.append(ds)

            else:
                print(f"   *****\n   Error while downloading the file {dataname}.csv\n   *****")
                
    else:
        print("No additional original dataset needs to be downloaded.")
        return
    
    
    # Export preparation
    name_pattern = re.compile("(.*)\d{4,6}-\d{4,6}")
    basename = name_pattern.findall(dataname)
    if len(basename) > 0:
        basename = basename[0]
    else:
        basename = 'QUOT_SIM2_'
    outpath = os.path.join(dst_folder, basename + f"{first_year}-{last_year}")
    
    
    if extension == '.csv':
# =============================================================================
#         if mask is not None:
#             print(f"Warning: As `extension` for output is {extension}, the `mask` will not be taken into account.")
#             print("If you want to save the results clipped over the specified mask, please pass `extension = '.nc'`")
# =============================================================================
        
        # ---- Clip data
        clipped_df = []
        
        if mask is not None:
            maskname = os.path.splitext(os.path.split(mask)[-1])[0]
            
            print(f"   _ clipping on {maskname}")
            
            i = 0
            for d in data_to_merge:
                i += 1
                print(f"      . {i}/{len(data_to_merge)}")
                
                clipped_df.append(geo.clip(d, src_crs = 27572, mask = mask))
        
        else:
            clipped_df = data_to_merge
        
        # ---- Merge pandas DataFrames
        print("Merging...")
        merged_df = geo.merge_data(clipped_df)
        
        # Revert temporary corrections on x and y dimension
        merged_df[['x', 'y']] = merged_df[['x', 'y']]/100 # convert hm to m
        merged_df.rename(columns = {'x': 'LAMBX', 'y': 'LAMBY', 'time': 'DATE'}, inplace = True)
        
        # ---- Export
        print("\nExporting...")
        print(f"   _ to {outpath}.csv")
        merged_df.to_csv(outpath + '.csv',
                         sep = ';',
                         header = True,
                         decimal = '.')
        
    elif extension == '.nc':
        # Clip and merge files in folder
        
# =============================================================================
#         if 'PRETOT' in data_to_merge[0]:
#             var_list = var_list + ['PRETOT']
# =============================================================================
        
        for var in data_to_merge[0].data_vars: 
            print(f"\nProcessing {var}...")

            # ---- Clip data
            clipped_ds = []
            
            if mask is not None:
                maskname = os.path.splitext(os.path.split(mask)[-1])[0]
                mask_ds = geo.load_any(mask)
                if isinstance(mask_ds, xr.Dataset):
                    mask_ds = mask_ds.where(mask_ds > 0, drop = True)
                    bounds = mask_ds.rio.bounds()
                    try: mask_crs = mask_ds.rio.crs
                    except: mask_crs = None
                elif isinstance(mask_ds, gpd.GeoDataFrame):
                    bounds = mask_ds.total_bounds
                    try: mask_crs = mask_ds.crs
                    except: mask_crs = None
                
                print(f"   _ clipping on {maskname}")
                
                i = 0
                for d in data_to_merge:
                    i += 1
                    print(f"      . {i}/{len(data_to_merge)}")
                    
                    clipped_ds.append(geo.clip(d[[var]], 
                                               bounds = bounds,
                                               bounds_crs = mask_crs))
            
            else:
                clipped_ds = data_to_merge
            
            # ---- Merge netCDF files
            # Add local files
            _, varfilelist = geo.get_filelist(dst_folder, 
                                              extension = ".nc", 
                                              tag = '^' + var)
            # it is necessary to add '^' before the string so that it will detect
            # only the files that start with var. To avoid confusion between
            # 'T_Q' and 'PRETOT_Q' (which also contains 'T_Q')
            for localfile in varfilelist:
                clipped_ds.append(geo.load_any(os.path.join(dst_folder, localfile)))
            
            print(f"   _ merging {len(clipped_ds)} files")
            merged_ds = geo.merge_data(clipped_ds, update_val = True)
            # update_val = True because sometimes SIM2 original datasets can overlap
            for localfile in varfilelist:
                os.remove(os.path.join(dst_folder, localfile))
            
            # ---- Export
            print("\nExporting...")
            var_outpath = os.path.join(os.path.split(outpath)[0],
                         f"{var}_" +  os.path.split(outpath)[-1]) + ".nc"
            print(f"   _ to {var_outpath}")

            geo.export(merged_ds, var_outpath)






# -*- coding: utf-8 -*-
"""
Created on Fri Jan 31 2025

@author: Alexandre Coche
"""

#%% Imports
import os
import geohydroconvert as ghc


#%% Pipeline
def pipeline(data, *, 
             mask=None, bounds=None, 
             resolution=None, x0=None, y0=None,
             base_template=None, **rio_kwargs,
             ):
    """
    Pipeline to execute all the necessary steps to format ERA5-Land original
    datasets to CWatM.

    Parameters
    ----------
    data : str, list of str
        Folder, filepath or list of filepaths
    mask : optional
        Filepath of geopandas.dataframe of mask.
        The default is None.
    bounds : TYPE, optional
        DESCRIPTION. The default is None.
    resolution : TYPE, optional
        DESCRIPTION. The default is None.
    x0: number, optional
        Origin of the X-axis, used to align the reprojection grid. 
        The default is None
    y0: number, optional
        Origin of the Y-axis, used to align the reprojection grid. 
        The default is None
    base_template : str, optional
        Filepath, used as a template for spatial profile. The default is None.
    
    **rio_kwargs : keyword args
        Argument passed to the xarray.Dataset.rio.reproject() function call.
        These arguments are prioritary over base_template attributes.
        May contain: 
            . dst_crs
            . resolution
            . shape
            . transform
            . resampling
            . nodata
            > help(xr.Dataset.rio.reproject)

    Returns
    -------
    None. Generates the formated files.

    """
    #### Initialisation
    # -+-+-+-+-+-+-+-+-   
    if isinstance(resolution, int) or resolution is None:
        resolution = [resolution]
        
    if not isinstance(mask, list):
        mask = [mask]
    
    ### Pipeline
    # -+-+-+-+-+-
    # ---- Convertit l'argument 'data' en liste de fichiers
    data_folder, filelist = ghc.get_filelist(data, filetype = '.nc')

    outpath = os.path.splitext(filelist[0])[0]
    outpath = os.path.join(data_folder, outpath)
    
    
    # ---- Corrige les données
    #   - fusionne les fichiers
    #   - convertit valeurs horaires en journalières
    #   - corrige les unités si besoin
    ds = ghc.convert_to_cwatm([os.path.join(data_folder, f) for f in filelist])
    
    
    # ---- Géoréférence les données
    #   - embed the coordinate reference system coord
    #   - standardize coords, to be recognized by QGIS
    ds = ghc.georef(data = ds, dst_crs = 4326)
    
    
    # ---- Reprojections et exports
    for res in resolution:
        if res is None:
            res_suffix = ''
            if 'resolution' in rio_kwargs:
                rio_kwargs.pop('resolution')
        else:
            res_suffix = f'_{res}m'
            rio_kwargs['resolution'] = res
        n_mask = 0
        for m in mask:
            n_mask += 1
            ds_rprj = ghc.reproject(data = ds, mask = m, bounds = bounds,
                                 x0 = x0, y0 = y0, base_template = base_template,
                                 **rio_kwargs)
            ghc.export(ds_rprj, outpath + res_suffix + f'_mask{n_mask}' + '.nc')
            

#%% Correct biaises
def correct_biais(data, correct_factor, to_dailysum, progressive):
    """
    Correction très simpliste des valeurs ERA5-Land en appliquant un facteur 
    de correction.
    
    Fonction plutôt à utiliser sur les données finales formatées
    correct_factor:
        - for precipitations (hourly) : 1.8
        - for precipitations (daily sum): 0.087 
        - for radiations (daily sum): 0.0715 (and progressive = True)
        - for potential evapotranspiration pos (daily sum): 0.04
    """
    
    ghc.correct_era5_bias(data, correct_factor, to_dailysum, progressive)


#%% Additional variables
def derived_var(data):
    """
    To generate the secondary data from ERA5-Land data, such as:
        - crop and water standard ETP from pan evaporation
        - wind speed from U- and V-components
        - relative humidity from T°, P and dewpoint
        
    Parameters
    ----------
    data : filepath or xarray.dataset (or xarray.dataarray)
        Main dataset used to generate secondary quantities.
    
    Returns
    -------
    None. Generates intended files.
    """
    
    ghc.secondary_climvar(data)
